from enum import Enum, auto


class ErrorPolicy(Enum):
    """
    THROW_EXTRACTION: throw error extraction and return
    THROW_DOCUMENT: throw whole document
    PROCESS: still process by converting to processable type
    RAISE: raise error
    """
    THROW_EXTRACTION = auto()
    THROW_DOCUMENT = auto()
    PROCESS = auto()
    RAISE = auto()


class ETKValueError(Exception):
    pass


class KgValueError(ETKValueError):
    def __init__(self, message=""):
        super(KgValueError, self).__init__(message)


class ExtractorValueError(ETKValueError):
    def __init__(self, message=""):
        super(ExtractorValueError, self).__init__(message)


class TokenizerValueError(ETKValueError):
    def __init__(self, message=""):
        super(TokenizerValueError, self).__init__(message)


class ISODateError(Exception):
    def __init__(self, message=""):
        super(ISODateError, self).__init__(message)


class InvalidJsonPathError(Exception):
    def __init__(self, message=""):
        super(InvalidJsonPathError, self).__init__(message)


class StoreExtractionError(Exception):
    def __init__(self, message=""):
        super(StoreExtractionError, self).__init__(message)


class NotGetETKModuleError(Exception):
    def __init__(self, message=""):
        super(NotGetETKModuleError, self).__init__(message)


class InvalidArgumentsError(Exception):
    def __init__(self, message=""):
        super(InvalidArgumentsError, self).__init__(message)


class InvalidFilePathError(Exception):
    def __init__(self, message=""):
        super(InvalidFilePathError, self).__init__(message)


class ExtractorError(Exception):
    def __init__(self, message=""):
        super(ExtractorError, self).__init__(message)


class UndefinedFieldError(Exception):
    def __init__(self, message=""):
        super(UndefinedFieldError, self).__init__(message)



import unittest
from etk.extraction import Extractable
from etk.tokenizer import Tokenizer


class TestExtractable(unittest.TestCase):

    def test_Extractable(self) -> None:
        e = Extractable({'extracted_value': [{1: 2, 'das': [1, 2, 3]}], 'confidence': 2.3})
        t = Tokenizer(keep_multi_space=False)
        tokens = e.get_tokens(t)
        token_attrs = []
        for i in tokens:
            token_attrs.append({"orth": i.orth_, "offset": i.idx, "full_shape": i._.full_shape})
        expected_token = [
            {'orth': 'extracted', 'offset': 0, 'full_shape': 'xxxxxxxxx'},
            {'orth': '_', 'offset': 9, 'full_shape': '_'},
            {'orth': 'value', 'offset': 10, 'full_shape': 'xxxxx'},
            {'orth': ':', 'offset': 16, 'full_shape': ':'},
            {'orth': '1', 'offset': 18, 'full_shape': 'd'},
            {'orth': ':', 'offset': 20, 'full_shape': ':'},
            {'orth': '2', 'offset': 22, 'full_shape': 'd'},
            {'orth': 'das', 'offset': 24, 'full_shape': 'xxx'},
            {'orth': ':', 'offset': 28, 'full_shape': ':'},
            {'orth': '1', 'offset': 30, 'full_shape': 'd'},
            {'orth': '2', 'offset': 32, 'full_shape': 'd'},
            {'orth': '3', 'offset': 34, 'full_shape': 'd'},
            {'orth': 'confidence', 'offset': 36, 'full_shape': 'xxxxxxxxxx'},
            {'orth': ':', 'offset': 47, 'full_shape': ':'},
            {'orth': '2.3', 'offset': 49, 'full_shape': 'd.d'}
        ]

        self.assertEqual(token_attrs, expected_token)
        text = e.get_string()
        expected_str = "extracted_value : 1 : 2 das : 1 2 3    confidence : 2.3 "

        self.assertEqual(text, expected_str)

if __name__ == '__main__':
    unittest.main()


from etk.extraction import Extractable
from etk.origin_record import OriginRecord


class KnowledgeGraphProvenanceRecord(Extractable):
    """
    A Provenance Record containing details of Extraction Results history.
    """

    def __init__(self, _id, _type: str, reference_type: str, _value: str, json_path: str, doc) -> None:

        Extractable.__init__(self)
        self.id = _id
        self._type = _type
        self._value = _value
        self.json_path = json_path
        self.reference_type = reference_type
        self.origin_doc = doc

    def get_origins(self, value):
        if self.reference_type == 'constant':
            return []
        else:
            origins = []
            break_point = self.json_path.rfind('.')
            if break_point == -1:
                json_path = self.json_path
            else:
                json_path = self.json_path[0:break_point]
            if json_path in self.origin_doc._jsonpath_provenances:
                matching_provenance_ids = self.origin_doc._jsonpath_provenances[json_path]
                for matched_prov_id in matching_provenance_ids:
                    prov = self.origin_doc._provenances[matched_prov_id]
                    origins.extend(prov.get_origins(value))
            else:
                origin_record = OriginRecord(json_path, None, None, self.origin_doc)
                origins.extend(origin_record)
            return origins



import spacy
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from typing import List


class SpacyNerExtractor(Extractor):
    """
    **Description**
        This extractor takes a list of spaCy NER tag as reference, and extract
        the tag matched substring from the input text

    Examples:
        ::

            get_attr = ['PERSON', 'ORG', 'GPE']
            spacy_ner_extractor = SpacyNerExtractor()
            spacy_ner_extractor.extract(text=text, get_attr=get_attr)

    """
    def __init__(self, extractor_name: str, nlp=spacy.load('en_core_web_sm')):
        Extractor.__init__(self, input_type=InputType.TEXT,
                           category="built_in_extractor",
                           name=extractor_name)
        self.__nlp = nlp

    # all_attrs = ['PERSON', 'NORP', 'FAC', 'ORG', 'GPE', 'LOC', 'PRODUCT', 'EVENT', 'WORK_OF_ART', 'LAW', 'LANGUAGE',
    #              'DATE', 'TIME', 'PERCENT', 'MONEY', 'QUANTITY', 'ORDINAL', 'CARDINAL']
    def extract(self, text: str, get_attr=['PERSON', 'ORG', 'GPE']) -> List[Extraction]:
        """
        Args:
            text (str): the text to extract from.
            get_attr (List[str]): The spaCy NER attributes we're interested in.

        Returns:
            List(Extraction): the list of extraction or the empty list if there are no matches.
        """
        doc = self.__nlp(text)
        attr_list = list()
        for ent in doc.ents:
            if ent.label_ in get_attr:
                attr_list.append(Extraction(extractor_name=self.name,
                                            start_char=int(ent.start_char),
                                            end_char=int(ent.end_char),
                                            value=ent.text,
                                            tag=ent.label_,
                                            start_token=ent.start,
                                            end_token=ent.end))
        return attr_list



import unittest
from etk.extractors.glossary_extractor import GlossaryExtractor
from etk.tokenizer import Tokenizer
from etk.etk import ETK
from etk.document import Document
import time


class TestGlossaryExtractor(unittest.TestCase):
    def setUp(self):
        self.glossary_1 = ['Beijing', 'Los Angeles', 'New York', 'Shanghai']
        self.glossary_2 = ['Beijing', 'Los Angeles', 'New York City', 'Shanghai']

    def test_glossary_extractor(self) -> None:
        t = Tokenizer()
        text = 'i live in los angeles. my hometown is Beijing. I love New York City.'
        tokens = t.tokenize(text)

        ge = GlossaryExtractor(self.glossary_1, 'test_glossary', t, 3, False)
        results = [i.value for i in ge.extract(tokens)]
        expected = ['Beijing', 'los angeles', 'New York']

        self.assertEqual(results, expected)

    def test_case_sensitive(self) -> None:
        t = Tokenizer()
        text = 'i live in los angeles. my hometown is Beijing. I love New York City.'
        tokens = t.tokenize(text)

        ge = GlossaryExtractor(self.glossary_1, 'test_glossary', t, 2, True)

        results = [i.value for i in ge.extract(tokens)]
        expected = ['Beijing', 'New York']

        self.assertEqual(results, expected)

    def test_n_grams(self) -> None:
        t = Tokenizer()
        text = 'i live in los angeles. my hometown is Beijing. I love New York City.'
        tokens = t.tokenize(text)

        ge = GlossaryExtractor(self.glossary_2, 'test_glossary', t, 2, False)

        results = [i.value for i in ge.extract(tokens)]
        expected = ['Beijing', 'los angeles']

        self.assertEqual(results, expected)

    def test_etk__spacy_glossary_extraction(self):
        etk = ETK(use_spacy_tokenizer=True)
        s = time.time()
        city_extractor = GlossaryExtractor(['los angeles', 'new york', 'angeles'], 'city_extractor',
                                           etk.default_tokenizer,
                                           case_sensitive=False, ngrams=3)
        doc_json = {'text': 'i live in los angeles. my hometown is Beijing. I love New York City.'}
        doc = Document(etk, cdr_document=doc_json, mime_type='json', url='', doc_id='1')
        t_segments = doc.select_segments("$.text")
        for t_segment in t_segments:
            extracted_cities = doc.extract(city_extractor, t_segment)
            for extracted_city in extracted_cities:
                self.assertTrue(extracted_city.value in ['los angeles', 'New York', 'angeles'])

    def test_etk_crf_glossary_extraction(self):
        etk = ETK(use_spacy_tokenizer=False)
        s = time.time()
        city_extractor = GlossaryExtractor(['los angeles', 'new york', 'angeles'], 'city_extractor',
                                           etk.default_tokenizer,
                                           case_sensitive=False, ngrams=3)
        doc_json = {'text': 'i live in los angeles. my hometown is Beijing. I love New York City.'}
        doc = Document(etk, cdr_document=doc_json, mime_type='json', url='', doc_id='1')
        t_segments = doc.select_segments("$.text")
        for t_segment in t_segments:
            extracted_cities = doc.extract(city_extractor, t_segment)
            for extracted_city in extracted_cities:
                self.assertTrue(extracted_city.value in ['los angeles', 'New York', 'angeles'])


if __name__ == '__main__':
    unittest.main()



from warnings import warn
from typing import List
from enum import Enum, auto
import datetime, re, calendar, pytz
from tzlocal import get_localzone
from dateutil.relativedelta import relativedelta
from langdetect import detect

from etk.etk import ETK
from etk.extractor import Extractor, InputType
from etk.extractors.spacy_rule_extractor import SpacyRuleExtractor
from etk.extraction import Extraction
from etk.dependencies.date_extractor_resources.date_regex_generator import DateRegexGenerator
from etk.dependencies.date_extractor_resources.constants import units, singleton_regex, \
    spacy_rules, directions, num_to_digit, foreign_to_english, language_date_order, \
    day_of_week_to_number, illegal, possible_illegal, possible_illegal_3

# to avoid typo:
EXTRACT_FIRST_DATE_ONLY = 'extract_first_date_only'
ADDITIONAL_FORMATS = 'additional_formats'
USE_DEFAULT_FORMATS = 'use_default_formats'
IGNORE_DATES_BEFORE = 'ignore_dates_before'
IGNORE_DATES_AFTER = 'ignore_dates_after'
DETECT_RELATIVE_DATES = 'detect_relative_dates'
RELATIVE_BASE = 'relative_base'
PREFERRED_DATE_ORDER = 'preferred_date_order'
PREFER_LANGUAGE_DATE_ORDER = 'prefer_language_date_order'
TIMEZONE = 'timezone'
TO_TIMEZONE = 'to_timezone'
RETURN_AS_TIMEZONE_AWARE = 'return_as_timezone_aware'
PREFER_DAY_OF_MONTH = 'prefer_day_of_month'
PREFER_DATES_FROM = 'prefer_dates_from'
DATE_VALUE_RESOLUTION = 'date_value_resolution'
MIN_RESOLUTION = 'min_resolution'


class DateResolution(Enum):
    """
    date resolution when convert a datetime object to iso format string
    """
    SECOND = auto()
    MINUTE = auto()
    HOUR = auto()
    DAY = auto()
    MONTH = auto()
    YEAR = auto()
    ORIGINAL = auto() # keep original resolution


class DateResolutionHelper():

    _sorted_resolution = [DateResolution.SECOND, DateResolution.MINUTE, DateResolution.HOUR,
                          DateResolution.DAY, DateResolution.MONTH, DateResolution.YEAR]
    _pattern_resolution_map = {
        '%a': DateResolution.DAY,
        '%A': DateResolution.DAY,
        '%w': DateResolution.DAY,
        '%d': DateResolution.DAY,
        '%b': DateResolution.MONTH,
        '%B': DateResolution.MONTH,
        '%m': DateResolution.MONTH,
        '%y': DateResolution.YEAR,
        '%Y': DateResolution.YEAR,
        '%H': DateResolution.HOUR,
        '%I': DateResolution.HOUR,
        '%p': None, # am / pm
        '%M': DateResolution.MINUTE,
        '%S': DateResolution.SECOND,
        '%f': DateResolution.SECOND,
        '%z': None, # UTC offset
        '%Z': None, # timezone name
        '%j': DateResolution.DAY,
        '%U': DateResolution.DAY, # week
        '%W': DateResolution.DAY, # week
        '%c': None, # locale representation
        '%x': None,
        '%X': None,
        '%%': None,
        '%G': DateResolution.YEAR,
        '%u': DateResolution.DAY,
        '%V': DateResolution.DAY
    }

    @staticmethod
    def min_resolution(pattern: list):
        min_index = len(DateResolutionHelper._sorted_resolution)
        for p in pattern:
            p = DateResolutionHelper._pattern_resolution_map.get(p, None)
            if not p:
                continue
            min_index = min(DateResolutionHelper._sorted_resolution.index(p), min_index)
        return DateResolutionHelper._sorted_resolution[min_index]


class DateExtractor(Extractor):
    """
    **Description**
        This extractor pre-defines a rich set of datetime regexp to detect any format of
        timestamp from the input text. In addition, it employees the spaCy rule to infer
        the specific datetime from a relative datetime and relative datetime base,
        for instance, two days ago with relative datetime base 02/02/2018.

    Examples:
        ::

            date_extractor = (etk=self.etk)
            date_extractor.extract(text=input_doc,
                                extract_first_date_only=False,  # first valid
                                additional_formats=['%Y@%m@%d', '%a %Y, %b %d'],
                                use_default_formats=True,
                                ignore_dates_before=ignore_before,
                                ignore_dates_after=ignore_after,
                                relative_base=relative_base,
                                preferred_date_order="DMY",
                                prefer_language_date_order=True,
                                timezone='GMT',
                                to_timezone='UTC',
                                return_as_timezone_aware=False,
                                prefer_day_of_month='first',
                                prefer_dates_from='future',
                                )

    """

    def __init__(self, etk: ETK=None, extractor_name: str='date extractor') -> None:
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="data extractor",
                           name=extractor_name)

        # The 'final_regex' and 'symbol_list' are generated by 'DateRegexGenerator'
        # If the single regexes are changed or more patterns are added,
        # please re-generate 'final_regex' and 'symbol_list' and paste here.
        d = DateRegexGenerator(singleton_regex, units)
        self._final_regex = d.final_regex
        self._symbol_list = d.symbol_list
        self._settings = {}
        self._etk = etk
        self._lan = 'en'

    def extract(self, text: str = None,
                extract_first_date_only: bool = False,
                additional_formats: List[str] = list(),
                use_default_formats: bool = False,
                ignore_dates_before: datetime.datetime = None,
                ignore_dates_after: datetime.datetime = None,
                detect_relative_dates: bool = False,
                relative_base: datetime.datetime = None,
                preferred_date_order: str = "MDY",
                prefer_language_date_order: bool = True,
                timezone: str = None,
                to_timezone: str = None,
                return_as_timezone_aware: bool = True,
                prefer_day_of_month: str = "first",
                prefer_dates_from: str = "current",
                date_value_resolution: DateResolution = DateResolution.DAY,
                ) -> List[Extraction]:
        """
        Args:
            text (str):  extract dates from this 'text', default to None
            extract_first_date_only (bool): extract the first valid date only or extract all, default to False
            additional_formats (List[str]):  user defined formats for extraction, default to empty list
            use_default_formats (bool): if use default formats together with addtional_formats, default to False
            ignore_dates_before (datetime.datetime): ignore dates before 'ignore_dates_before', default to None
            ignore_dates_after (datetime.datetime): ignore dates after 'ignore_dates_after', default to None
            detect_relative_dates (bool): if detect relative dates like '9 days before', default to False
            relative_base (datetime.datetime): offset relative dates detected based on 'relative_base', default to None
            preferred_date_order (enum['MDY', 'DMY', 'YMD']): preferred date order when ambiguous, default to 'MDY'
            prefer_language_date_order (bool): if use the text language's preferred order, default to True
            timezone (str): add 'timezone' if there is no timezone information in the extracted date, default to None
            to_timezone (str): convert all dates extracted to this timezone, default to None
            return_as_timezone_aware (bool): returned datetime timezone awareness, default to None
            prefer_day_of_month (enum['first', 'current', 'last']): use which day of the month when there is no 'day', default to 'first'
            prefer_dates_from (enum['past', 'current', 'future']): use which date when there is few info(e.g. only month), default to 'current'
            date_value_resolution (enum[DateResolution.SECOND, DateResolution.MINUTE, DateResolution.HOUR, \
                DateResolution.DAY, DateResolution.MONTH, DateResolution.YEAR]): specify resolution \
                when convert to iso format string, default to DateResolution.DAY

        Returns:
            List[Extraction]: List of extractions, the information including::

                Extraction._value: iso format string,
                Extraction._provenance: provenance information including:
                {
                    'start_char': int - start_char,
                    'end_char': int - end_char
                },
                Extraction._addition_inf: additional information including:
                {
                    'date_object': datetime.datetime - the datetime object,
                    'original_text': str - the original str extracted from text,
                    'language': enum['en', 'es'] - language of the date
                }
        """

        if return_as_timezone_aware:
            self._default_tz = pytz.timezone(timezone) if timezone else get_localzone()
            if ignore_dates_before and not ignore_dates_before.tzinfo:
                ignore_dates_before = ignore_dates_before.astimezone(self._default_tz)
            if ignore_dates_after and not ignore_dates_after.tzinfo:
                ignore_dates_after = ignore_dates_after.astimezone(self._default_tz)
            if relative_base and not relative_base.tzinfo:
                relative_base = relative_base.astimezone(self._default_tz)
        else:
            if ignore_dates_before and ignore_dates_before.tzinfo:
                ignore_dates_before = ignore_dates_before.replace(tzinfo=None)
            if ignore_dates_after and ignore_dates_after.tzinfo:
                ignore_dates_after = ignore_dates_after.replace(tzinfo=None)
            if relative_base and relative_base.tzinfo:
                relative_base = relative_base.replace(tzinfo=None)

        if prefer_language_date_order:
            try:
                self._lan = detect(text)
            except Exception as e:
                warn('DateExtractor: Catch LangDetectException ' + str(e))
                warn(message='DateExtractor: Catch LangDetectException {}'.format(str(e)))

        self._settings = {
            EXTRACT_FIRST_DATE_ONLY: extract_first_date_only,
            ADDITIONAL_FORMATS: additional_formats,
            USE_DEFAULT_FORMATS: use_default_formats,
            IGNORE_DATES_BEFORE: ignore_dates_before,
            IGNORE_DATES_AFTER: ignore_dates_after,
            DETECT_RELATIVE_DATES: detect_relative_dates,
            RELATIVE_BASE: relative_base,
            PREFERRED_DATE_ORDER: preferred_date_order,
            PREFER_LANGUAGE_DATE_ORDER: prefer_language_date_order,
            TIMEZONE: timezone,
            TO_TIMEZONE: to_timezone,
            RETURN_AS_TIMEZONE_AWARE: return_as_timezone_aware,
            PREFER_DAY_OF_MONTH: prefer_day_of_month,
            PREFER_DATES_FROM: prefer_dates_from,
            DATE_VALUE_RESOLUTION: date_value_resolution
        }

        results = []
        additional_regex = []
        if additional_formats:
            for date_format in additional_formats:
                order = ''
                reg = date_format
                for key in singleton_regex:
                    if key[0] == '%':
                        reg2 = re.sub(key, singleton_regex[key], reg)
                        if reg != reg2:
                            if key in units['M']:
                                order += 'M'
                            elif key in units['Y']:
                                order += 'Y'
                            elif key in units['D']:
                                order += 'D'
                            reg = reg2
                additional_regex.append({
                    'reg': reg,
                    'pattern': date_format,
                    'order': order,
                })
            for r in additional_regex:
                try:
                    matches = [self._wrap_date_match(r['order'], match, pattern=r['pattern']) for
                           match in re.finditer(r['reg'], text, re.I) if match]
                    if matches:
                        results.append(matches)
                except:
                    warn('DateExtractor: Failed to extract with additional format ' + str(r) + '.')
            if use_default_formats:
                for order in self._final_regex.keys():
                    matches = [self._wrap_date_match(order, match) for match
                               in re.finditer(self._final_regex[order], text, re.I) if match]
                    if matches:
                        results.append(matches)
        else:
            for order in self._final_regex.keys():
                matches = [self._wrap_date_match(order, match) for match
                           in re.finditer(self._final_regex[order], text, re.I) if match]
                results.append(matches)

        # for absolute dates:
        ans = self._remove_overlapped_date_str(results)
        # for relative dates:
        if detect_relative_dates:
            ans += self._extract_relative_dates(text)

        return ans

    def _wrap_extraction(self, date_object: datetime.datetime,
                        original_text: str,
                        start_char: int,
                        end_char: int
                        ) -> Extraction or None:
        """
        wrap the final result as an Extraction and return

        """
        try:
            resolution = self._settings[MIN_RESOLUTION] \
                    if self._settings[DATE_VALUE_RESOLUTION] == DateResolution.ORIGINAL \
                    else self._settings[DATE_VALUE_RESOLUTION]
            e = Extraction(self._convert_to_iso_format(date_object, resolution=resolution),
                           start_char=start_char,
                           end_char=end_char,
                           extractor_name=self._name,
                           date_object=date_object,
                           original_date=original_text)
            return e
        except Exception as e:
            warn('DateExtractor: Failed to wrap result ' + str(original_text) + ' with Extraction class.\n'
                                                                                'Catch ' + str(e))
            return None

    def _remove_overlapped_date_str(self, results: List[List[dict]]) -> List[Extraction]:
        """
        some string may be matched by multiple date templates,
        deduplicate the results and return a single list

        """
        res = []
        all_results = []
        for x in results:
            all_results = all_results + x
        if not all_results or len(all_results) == 0:
            return list()
        all_results.sort(key=lambda k: k['start'])
        cur_max = None
        i = 0
        while i < len(all_results) and not cur_max:
            if self._post_check(all_results[i]):
                cur_max = all_results[i]
            i += 1

        if not cur_max:
            return res

        while i < len(all_results):
            x = all_results[i]
            i += 1
            if not self._post_check(x):
                continue
            if cur_max['end'] <= x['start']:
                parsed_date = self._parse_date(cur_max)
                if parsed_date:
                    if self._settings[EXTRACT_FIRST_DATE_ONLY]:
                        return res
                    res.append(parsed_date)
                cur_max = x
            else:
                if len(x['value']) > len(cur_max['value']):
                    cur_max = x
                elif len(x['value']) == len(cur_max['value']):
                    if x['order'] in ['SINGLE_YEAR']:
                        cur_max = x
                    elif len(x['order']) == len(cur_max['order']):
                        if len(x['groups']) < len(cur_max['groups']):
                            cur_max = x
                        elif len(x['groups']) == len(cur_max['groups']):
                            if sum(ele is not None for ele in x['groups']) < sum(ele is not None for ele in cur_max['groups']):
                                cur_max = x
                            elif self._settings[PREFER_LANGUAGE_DATE_ORDER] and self._lan in language_date_order:
                                if x['order'] == language_date_order[self._lan]:
                                    cur_max = x
                                elif x['order'] == self._settings[PREFERRED_DATE_ORDER]:
                                    cur_max = x
                            elif x['order'] == self._settings[PREFERRED_DATE_ORDER]:
                                cur_max = x
        parsed_date = self._parse_date(cur_max)
        if parsed_date:
            if self._settings[EXTRACT_FIRST_DATE_ONLY]:
                return res
            res.append(parsed_date)
        return res

    def _parse_date(self, date_info: dict) -> Extraction or None:
        """
        parse a date string extracted to a datetime.datetime object
        apply the customizations like date range, date completion etc.

        """

        user_defined_pattern = None

        if date_info['pattern']:
            user_defined_pattern = re.findall(r'%[a-zA-Z]', date_info['pattern'])

        miss_day = miss_month = miss_year = miss_week = True

        i = 0
        pattern = list()
        formatted = list()

        for s in date_info['groups']:
            if s:
                p = self._symbol_list[date_info['order']][i] if not user_defined_pattern else user_defined_pattern[i]
                if p in units['D']:
                    miss_day = False
                elif p in units['M']:
                    miss_month = False
                elif p in units['Y']:
                    miss_year = False
                elif p in units['W']:
                    miss_week = False
                formatted_str = s.strip('.').strip().lower()
                if p in ['%B', '%b', '%A', '%a']:
                    if formatted_str in foreign_to_english:
                        # TODO: rearrange language detection in a better way
                        if self._lan == 'en':
                            continue
                        formatted_str = foreign_to_english[formatted_str]
                if p in ['%b', '%a']:
                    formatted_str = formatted_str[:3]
                formatted.append(re.sub(r'[^0-9+\-]', '', formatted_str) if p == '%z' else formatted_str)
                pattern.append(p)
            i += 1

        # TODO: deduplicate in the regex extraction part would be better
        exist, new_formatted, new_pattern = set(), [], []
        for i in range(len(pattern)):
            if pattern[i] not in exist:
                if re.match(r'[a-zA-Z]', formatted[i]) and pattern[i] == '%a':
                    miss_week = True
                else:
                    new_pattern.append(pattern[i])
                    new_formatted.append(formatted[i])
                    exist.add(pattern[i])
        formatted, pattern = new_formatted, new_pattern

        if formatted and pattern:
            try:
                if self._settings[DATE_VALUE_RESOLUTION] == DateResolution.ORIGINAL:
                    self._settings[MIN_RESOLUTION] = DateResolutionHelper.min_resolution(pattern)
                date = datetime.datetime.strptime('-'.join(formatted), '-'.join(pattern))
            except ValueError:
                try:
                    date = datetime.datetime.strptime('-'.join(formatted[:-1]), '-'.join(pattern[:-1]))
                except ValueError:
                    warn('DateExtractor: Failed to parse string to datetime object. \n'
                         'Patterns are not matched with string or the formats are not supported. ' +
                         '-'.join(formatted) + ' with ' + '-'.join(pattern))
                    return None

            if miss_year and miss_month and miss_day:
                today = datetime.datetime.now()
                if miss_week:
                    date = date.replace(day=today.day, month=today.month, year=today.year)
                else:
                    date = today
                    week_of_day = formatted[0].strip().lower()
                    if week_of_day in foreign_to_english:
                        week_of_day = foreign_to_english[week_of_day]
                    target = day_of_week_to_number[week_of_day] if week_of_day in day_of_week_to_number \
                        else today.weekday()
                    if self._settings[PREFER_DATES_FROM] == 'past':
                        date = date + relativedelta(days=-(date.weekday()+7-target)%7)
                    elif self._settings[PREFER_DATES_FROM] == 'future':
                        date = date + relativedelta(days=(target+7-date.weekday())%7)
                    else:
                        delta = target - date.weekday()
                        if abs(delta) <= 3:
                            date = date + relativedelta(days=delta)
                        else:
                            date = date + relativedelta(days=delta-7)

            else:
                if miss_day:
                    last = calendar.monthrange(date.year, date.month)[1]
                    if self._settings[PREFER_DAY_OF_MONTH] == 'current':
                        cur = datetime.datetime.now().day
                        date = date.replace(day=cur if cur <= last else last)
                    elif self._settings[PREFER_DAY_OF_MONTH] == 'last':
                        date = date.replace(day=last)

                if miss_year:
                    today = datetime.datetime.now()
                    date = date.replace(year=today.year)
                    next_year = date.replace(year=today.year+1)
                    last_year = date.replace(year=today.year-1)
                    if self._settings[PREFER_DATES_FROM] == 'past':
                        date = last_year if date > today else date
                    elif self._settings[PREFER_DATES_FROM] == 'future':
                        date = next_year if date < today else date
                    else:
                        if date > today and (date-today > today-last_year):
                            date = last_year
                        elif date < today and (today-date > next_year-today):
                            date = next_year

            date = self._post_process_date(date)

            if date:
                return self._wrap_extraction(date, date_info['value'], date_info['start'], date_info['end'])
        return None

    def _post_process_date(self, date: datetime.datetime) -> datetime.datetime or None:
        """

        apply date range and timezone conversion

        """
        if not date.tzinfo and self._settings[RETURN_AS_TIMEZONE_AWARE]:
            # cannot set time zone for time before 1883-11-19
            try:
                date = date.astimezone(self._default_tz)
            except ValueError as e:
                warn('DateExtractor: Failed to set timezone as ' + str(self._default_tz) + '. Catch ' + str(e))
        elif not self._settings[RETURN_AS_TIMEZONE_AWARE]:
            date = date.replace(tzinfo=None)

        try:
            if (self._settings[IGNORE_DATES_BEFORE] and date < self._settings[IGNORE_DATES_BEFORE]) or \
                    (self._settings[IGNORE_DATES_AFTER] and date > self._settings[IGNORE_DATES_AFTER]):
                return None
        except Exception as e:
            warn('DateExtractor: Failed to compare dates. Catch ' + str(e))

        # TODO: support more timezones abbr. (Only support what pytz supports currently)
        if self._settings[TO_TIMEZONE] and self._settings[RETURN_AS_TIMEZONE_AWARE]:
            try:
                date = date.astimezone(pytz.timezone(self._settings[TO_TIMEZONE]))
            except Exception as e:
                warn('DateExtractor: Failed to set timezone as ' + str(self._settings[TIMEZONE]) + '. Catch ' + str(e))

        return date

    def _extract_relative_dates(self, text: str) -> List[Extraction]:
        """

        Extract relative dates using spaCy rules

        Args:
            text: str - the text to extract the relative date strings from

        Returns: List of Extraction(s)

        """
        if not text or not self._etk:
            return list()
        base = self._settings[RELATIVE_BASE] if self._settings[RELATIVE_BASE] else datetime.datetime.now()
        if not self._settings[RETURN_AS_TIMEZONE_AWARE]:
            base = base.replace(tzinfo=None)
        elif not base.tzinfo:
            base = base.astimezone(self._default_tz)
        res = SpacyRuleExtractor(self._etk.default_nlp, spacy_rules, 'relative_date_extractor').extract(text)
        ans = list()
        for relative_date in res:
            if relative_date.rule_id == 'direction_number_unit':
                direction, measure, unit = relative_date.value.split()
                measure = num_to_digit[measure.lower()]
            elif relative_date.rule_id == 'number_unit_direction':
                measure, unit, direction = relative_date.value.split()
                measure = num_to_digit[measure.lower()]
            elif relative_date.rule_id == 'direction_digit_unit':
                direction, measure, unit = relative_date.value.split()
            elif relative_date.rule_id == 'digit_unit_direction':
                measure, unit, direction = relative_date.value.split()
            elif relative_date.rule_id == 'direction_unit':
                direction, unit = relative_date.value.split()
                measure = '1'
            elif relative_date.rule_id == 'the_day':
                unit = 'days'
                key_ = relative_date.value.split()[-1].lower()
                if key_ == 'today':
                    direction = 'ago'
                    measure = '0'
                else:
                    direction = 'ago' if key_ == 'yesterday' else 'later'
                    measure = '1' if len(relative_date.value.split()) == 1 else '2'
            else:
                continue
            unit = unit if unit[-1] == 's' else unit+'s'
            direction = directions[direction.lower()] if direction.lower() in directions else '+'
            delta_args = {unit: int(direction+measure)}
            relative_delta = relativedelta(**delta_args)
            date = self._post_process_date(base+relative_delta)
            if date:
                extraction_date = self._wrap_extraction(date,
                                                       relative_date.value,
                                                       relative_date.provenance['start_char'],
                                                       relative_date.provenance['end_char'])
                if extraction_date:
                    ans.append(extraction_date)
        return ans

    @staticmethod
    def _post_check(date_info: dict) -> bool:
        """

        Post check the extracted date string to filter out some false positives

        Args:
            date_info: dict - includes the extracted string, matching groups, patterns etc.

        Returns: bool - if the date extracted is valid

        """
        if date_info['pattern']:
            return True
        # TODO: consider more context when extract dates?? (e.g. for 'may')
        if date_info['value'] == 'may' or re.match(illegal, date_info['value'])\
            or (re.match(possible_illegal, date_info['value']) and len([g for g in date_info['groups'] if g]) != 2) \
            or (re.match(possible_illegal_3, date_info['value']) and len([g for g in date_info['groups'] if g]) != 3) \
            or (re.match('^\b?[0-9]{4}\b?$', date_info['value']) and len([g for g in date_info['groups'] if g]) > 1):
            return False
        return True

    @staticmethod
    def _convert_to_iso_format(date: datetime.datetime, resolution: DateResolution = DateResolution.DAY) -> str or None:
        """

        Args:
            date: datetime.datetime - datetime object to convert
            resolution: resolution of the iso format date to return

        Returns: string of iso format date

        """
        # TODO: currently the resolution is specified by the user, should it be decided what we have extracted,
        # E.g.: like if only year exists, use DateResolution.YEAR as resolution
        try:
            if date:
                date_str = date.isoformat()
                length = len(date_str)
                if resolution == DateResolution.YEAR and length >= 4:
                    return date_str[:4]
                elif resolution == DateResolution.MONTH and length >= 7:
                    return date_str[:7]
                elif resolution == DateResolution.DAY and length >= 10:
                    return date_str[:10]
                elif resolution == DateResolution.HOUR and length >= 13:
                    return date_str[:13]
                elif resolution == DateResolution.MINUTE and length >= 16:
                    return date_str[:16]
                elif resolution == DateResolution.SECOND and length >= 19:
                    return date_str[:19]
                return date_str
        except Exception as e:
            warn('DateExtractor: Failed to convert {} to ISO format. Catch {}.'.format(date, str(e)))
            return None
        warn('DateExtractor: Failed to convert {} to ISO format.'.format(date))
        return None

    @staticmethod
    def _wrap_date_match(order: str, match: re.match, pattern: str=None) -> dict or None:
        """

        Args:
            order: enums['MDY', 'DMY', 'YMD'] - order of the date
            match: re.match - a regex match object
            pattern: str - if user defined the pattern, record it here

        Returns:

        """
        return {
            'value': match.group(),
            'groups': match.groups(),
            'start': match.start(),
            'end': match.end(),
            'order': order,
            'pattern': pattern
        } if match else None



from warnings import warn
from typing import List
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from etk.tokenizer import Tokenizer
from etk.etk_exceptions import ExtractorError
from spacy.tokens import Token
from pygtrie import CharTrie
from itertools import *
from functools import reduce
import re

disable_spacy = ['tagger', 'parser', 'ner']
# TODO have an elegant way of handling spacy tokens vs str tokens
class GlossaryExtractor(Extractor):
    """
    **Description**
        This class takes a list of glossary as reference, extract the matched ngrams string
        from the tokenized input string.

    Examples:
        ::

            glossary = ['Beijing', 'Los Angeles', 'New York', 'Shanghai']
            glossary_extractor = GlossaryExtractor(glossary=glossary,
                                                  ngrams=3,
                                                  case_sensitive=True)
            glossary_extractor.extract(tokens=Tokenizer(input_text))

    """

    def __init__(self,
                 glossary: List[str],
                 extractor_name: str,
                 tokenizer: None,
                 ngrams: int = 2,
                 case_sensitive=False) -> None:
        # if we set tokenizer as None, extractor will use regex to extract tokens to expedite the extraction
        Extractor.__init__(self,
                           input_type=InputType.TOKENS,
                           category="glossary",
                           name=extractor_name)
        self._case_sensitive = case_sensitive
        self._default_tokenizer = tokenizer
        if not ngrams:
            ngrams = 0
            for word in glossary:
                ngrams = max(ngrams, len(self._default_tokenizer.tokenize(word)))
        self._ngrams = min(ngrams, 5)
        self._joiner = " "
        self._glossary = self._populate_trie(glossary)

    def extract(self, tokens: List[Token]) -> List[Extraction]:
        """
        Extracts information from a string(TEXT) with the GlossaryExtractor instance

        Args:
            token (List[Token]): list of spaCy token to be processed.

        Returns:
            List[Extraction]: the list of extraction or the empty list if there are no matches.

        """
        results = list()

        if len(tokens) > 0:
            if self._case_sensitive:
                new_tokens = [x.orth_ if isinstance(x, Token) else x for x in tokens]
            else:
                new_tokens = [x.lower_ if isinstance(x, Token) else x.lower() for x in tokens]
        else:
            return results

        try:
            ngrams_iter = self._generate_ngrams_with_context(new_tokens)
            results.extend(map(lambda term: self._wrap_value_with_context(tokens, term[1], term[2]),
                               filter(lambda term: isinstance(term[0], str),
                                      map(lambda term: (self._glossary.get(term[0]), term[1], term[2]),
                                          map(lambda term: (
                                              self._combine_ngrams(term[0], self._joiner), term[1], term[2]),
                                              ngrams_iter)))))
        except Exception as e:
            raise ExtractorError('GlossaryExtractor: Failed to extract with ' + self.name + '. Catch ' + str(e) + '. ')
        return results

    def _generate_ngrams_with_context(self, tokens: List[Token]) -> chain:
        """Generates the 1-gram to n-grams tuples of the list of tokens"""
        chained_ngrams_iter = self._generate_ngrams_with_context_helper(iter(tokens), 1)
        for n in range(2, self._ngrams + 1):
            ngrams_iter = tee(tokens, n)
            for j in range(1, n):
                for k in range(j):
                    next(ngrams_iter[j], None)
            ngrams_iter_with_context = self._generate_ngrams_with_context_helper(zip(*ngrams_iter), n)
            chained_ngrams_iter = chain(chained_ngrams_iter, ngrams_iter_with_context)
        return chained_ngrams_iter

    def _populate_trie(self, values: List[str]) -> CharTrie:
        """Takes a list and inserts its elements into a new trie and returns it"""
        if self._default_tokenizer:
            return reduce(self._populate_trie_reducer, iter(values), CharTrie())
        return reduce(self._populate_trie_reducer_regex, iter(values), CharTrie())

    def _populate_trie_reducer(self, trie_accumulator=CharTrie(), value="") -> CharTrie:
        """Adds value to trie accumulator"""
        if self._case_sensitive:
            key = self._joiner.join([x.orth_ if isinstance(x, Token) else x for x in
                                     self._default_tokenizer.tokenize(value, disable=disable_spacy)])
        else:
            key = self._joiner.join([x.lower_ if isinstance(x, Token) else x.lower() for x in
                                     self._default_tokenizer.tokenize(value, disable=disable_spacy)])
        trie_accumulator[key] = value
        return trie_accumulator

    def _populate_trie_reducer_regex(self, trie_accumulator=CharTrie(), value="") -> CharTrie:
        """Adds value to trie accumulator"""
        regex = re.compile(r"[A-Za-z0-9]+|[^\w\s]|_")
        if self._case_sensitive:
            key = self._joiner.join([x for x in re.findall(regex, value)])
        else:
            key = self._joiner.join([x.lower() for x in re.findall(regex, value)])
        trie_accumulator[key] = value
        return trie_accumulator

    def _wrap_value_with_context(self, tokens: List[Token], start: int, end: int) -> Extraction:
        """Wraps the final result"""
        return Extraction(' '.join([x.orth_ if isinstance(x, Token) else x for x in tokens[start:end]]),
                          self.name,
                          start_token=start,
                          end_token=end,
                          start_char=tokens[start].idx if isinstance(tokens[start], Token) else -1,
                          end_char=tokens[end - 1].idx + len(tokens[end - 1].orth_) if isinstance(tokens[end - 1],
                                                                                                  Token) else -1
                          )

    @staticmethod
    def _generate_ngrams_with_context_helper(ngrams_iter: iter, ngrams_len: int) -> map:
        """Updates the end index"""
        return map(lambda term: (term[1], term[0], term[0] + ngrams_len), enumerate(ngrams_iter))

    @staticmethod
    def _combine_ngrams(ngrams, joiner) -> str:
        """Construct keys for checking in trie"""
        if isinstance(ngrams, str):
            return ngrams
        else:
            combined = joiner.join(ngrams)
            return combined



import unittest, json
from etk.knowledge_graph import KGSchema
from etk.etk import ETK
from etk.etk_exceptions import KgValueError
from datetime import date, datetime
from etk.ontology_api import Ontology
from etk.ontology_namespacemanager import DIG


class TestKnowledgeGraph(unittest.TestCase):
    def setUp(self):
        sample_doc = {
            "projects": [
                {
                    "name": "etk",
                    "description": "version 2 of etk, implemented by Runqi12 Shao, Dongyu Li, Sylvia lin, Amandeep and others.",
                    "members": [
                        "dongyu",
                        "amandeep",
                        "sylvia",
                        "Runqi12"
                    ],
                    "date": "2007-12-05",
                    "place": "columbus:georgia:united states:-84.98771:32.46098",
                    "s": "segment_test_1"
                },
                {
                    "name": "rltk",
                    "description": "record linkage toolkit, implemented by Pedro, Mayank, Yixiang and several students.",
                    "members": [
                        "mayank",
                        "yixiang"
                    ],
                    "date": ["2007-12-05T23:19:00"],
                    "cost": -3213.32,
                    "s": "segment_test_2"
                }
            ]
        }
        kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

        etk = ETK(kg_schema)
        self.doc = etk.create_document(sample_doc)

    def test_add_segment_kg(self) -> None:
        sample_doc = self.doc
        segments = sample_doc.select_segments("projects[*].s")
        sample_doc.kg.add_value("segment", segments)
        expected_segments = ["segment_test_1", "segment_test_2"]
        self.assertTrue(sample_doc.kg.value["segment"][0]["key"] in expected_segments)
        self.assertTrue(sample_doc.kg.value["segment"][1]["key"] in expected_segments)
        self.assertTrue('provenances' in sample_doc.value)
        provenances = sample_doc.value['provenances']
        self.assertTrue(len(provenances) == 2)
        self.assertTrue(provenances[0]['reference_type'] == 'location')

    def test_KnowledgeGraph(self) -> None:
        sample_doc = self.doc

        try:
            sample_doc.kg.add_value("developer", json_path="projects[*].members[*]")
        except KgValueError:
            pass

        try:
            sample_doc.kg.add_value("test_date", json_path="projects[*].date[*]")
        except KgValueError:
            pass

        try:
            sample_doc.kg.add_value("test_add_value_date",
                                    value=[date(2018, 3, 28), {}, datetime(2018, 3, 28, 1, 1, 1)])
        except KgValueError:
            pass

        try:
            sample_doc.kg.add_value("test_location", json_path="projects[*].place")
        except KgValueError:
            pass

        try:
            sample_doc.kg.add_value("test_non_empty", value="")
            sample_doc.kg.add_value("test_non_empty", value="non-empty")
            sample_doc.kg.add_value("test_empty", value="", keep_empty=True)
            sample_doc.kg.add_value("test_empty", value="empty", keep_empty=True)
        except KgValueError:
            pass

        expected_developers = [
            {
                "value": "dongyu",
                "key": "dongyu"
            },
            {
                "value": "amandeep",
                "key": "amandeep"
            },
            {
                "value": "sylvia",
                "key": "sylvia"
            },
            {
                "value": "Runqi12",
                "key": "runqi12"
            },
            {
                "value": "mayank",
                "key": "mayank"
            },
            {
                "value": "yixiang",
                "key": "yixiang"
            }
        ]

        expected_date = [
            {
                "value": "2007-12-05T00:00:00",
                "key": "2007-12-05T00:00:00"
            },
            {
                "value": "2007-12-05T23:19:00",
                "key": "2007-12-05T23:19:00"
            }
        ]

        expected_add_value_date = [
            {
                "value": "2018-03-28",
                "key": "2018-03-28"
            },
            {
                "value": "2018-03-28T01:01:01",
                "key": "2018-03-28T01:01:01"
            }
        ]

        expected_location = [
            {
                "value": "columbus:georgia:united states:-84.98771:32.46098",
                "key": "columbus:georgia:united states:-84.98771:32.46098"
            }
        ]

        expected_non_empty = [{"key": "non-empty", "value": "non-empty"}]
        expected_empty = [{"key": "", "value": ""}, {"key": "empty", "value": "empty"}]

        self.assertEqual(expected_developers, sample_doc.kg.value["developer"])
        self.assertEqual(expected_date, sample_doc.kg.value["test_date"])
        self.assertEqual(expected_location, sample_doc.kg.value["test_location"])
        self.assertEqual(expected_add_value_date, sample_doc.kg.value["test_add_value_date"])
        self.assertEqual(expected_non_empty, sample_doc.kg.value["test_non_empty"])
        self.assertEqual(expected_empty, sample_doc.kg.value["test_empty"])

    def test_add_value_empty(self):
        self.doc.kg.add_value('test_zero', 0.0)
        self.assertEqual(self.doc.kg.value['test_zero'][0]['value'], 0.0)


class TestKnowledgeGraphWithOntology(unittest.TestCase):
    def setUp(self):
        ontology_content = '''
                @prefix : <http://dig.isi.edu/ontologies/dig/> .
                @prefix owl: <http://www.w3.org/2002/07/owl#> .
                @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
                @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
                @prefix schema: <http://schema.org/> .
                @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
                :Person a owl:Class ;
                    rdfs:subClassOf :Actor, :Biological_Object ;
                    :common_properties :label, :title, :religion ; .
                :has_name a owl:DatatypeProperty ;
                    schema:domainIncludes :Person ;
                    schema:rangeIncludes xsd:string ; .
                :has_child a owl:ObjectProperty ;
                    schema:domainIncludes :Person ;
                    schema:rangeIncludes :Person ; .
            '''
        ontology = Ontology(ontology_content, validation=False, include_undefined_class=True, quiet=True)
        kg_schema = KGSchema(ontology.merge_with_master_config(dict()))
        etk = ETK(kg_schema=kg_schema, ontology=ontology, generate_json_ld=True)
        etk2 = ETK(kg_schema=kg_schema, ontology=ontology, generate_json_ld=False)
        self.doc = etk.create_document(dict(), doc_id='http://xxx/1', type_=[DIG.Person.toPython()])
        self.doc2 = etk2.create_document(dict(), doc_id='http://xxx/2', type_=[DIG.Person.toPython()])

    def test_valid_kg_jsonld(self):
        kg = self.doc.kg
        self.assertIn('@id', kg._kg)
        self.assertEqual('http://xxx/1', kg._kg['@id'])
        self.assertIn('@type', kg._kg)
        self.assertIn(DIG.Person.toPython(), kg._kg['@type'])

    def test_valid_kg(self):
        kg = self.doc2.kg
        self.assertNotIn('@id', kg._kg)
        self.assertNotIn('@type', kg._kg)

    def test_add_value_kg_jsonld(self):
        kg = self.doc.kg
        field_name = kg.context_resolve(DIG.has_name)
        self.assertEqual('has_name', field_name)
        kg.add_value(field_name, 'Jack')
        self.assertIn({'@value': 'Jack'}, kg._kg[field_name])
        field_child = kg.context_resolve(DIG.has_child)
        self.assertEqual('has_child', field_child)
        child1 = 'http://xxx/2'
        child2 = {'@id': 'http://xxx/3', 'has_name': 'Daniels', '@type': [DIG.Person],
                  '@context': {'has_name': DIG.has_name.toPython()}}
        kg.add_value(field_child, child1)
        kg.add_value(field_child, child2)
        self.assertIn({'@id': 'http://xxx/2'}, kg._kg[field_child])

    def test_add_value_kg(self):
        kg = self.doc2.kg

        field_name = kg.context_resolve(DIG.has_name)

        self.assertEqual('has_name', field_name)
        kg.add_value(field_name, 'Jack')
        self.assertIn({'value': 'Jack', "key": "jack"}, kg._kg[field_name])



# -*- coding: utf-8 -*-
from typing import List
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction

from bs4 import BeautifulSoup
from bs4.element import Comment
import re


class Toolkit:
    @staticmethod
    def create_table_array(t, put_extractions=False):
        rows = t['rows']
        tt = []
        max_cols = t['features']['max_cols_in_a_row']
        for r in rows:
            new_r = ['' for xx in range(max_cols)]
            for i, c in enumerate(r['cells']):
                text = c['text']
                text = text.lower()
                text = text.strip()
                if put_extractions and 'data_extraction' in c:
                    data_extractions = c['data_extraction']
                    for key in data_extractions.keys():
                        text += ' DUMMY' + key.upper()
                new_r[i] = text.strip()
            tt.append(new_r)
        return tt

    @staticmethod
    def regulize_cells(t):
        for r in t:
            for i in range(len(r)):
                r[i] = re.sub('[0-9]', 'NUM', r[i])
                # for x in re.findall('([0-9])', r[i]):
                #     int_x = int(x)
                #     if int_x < 5:
                #         r[i] = re.sub(x, 'SSSS', r[i])
                #     else:
                #         r[i] = re.sub(x, 'LLLL', r[i])
                for x in re.findall('([a-z][a-z][a-z]+@)', r[i]):
                    r[i] = re.sub(x, 'EMAILNAME ', r[i])

    @staticmethod
    def clean_cells(t): # modifies t
        for r in t:
            for i in range(len(r)):
                r[i] = re.sub(r'[^\x00-\x7F]',' ', r[i]) #remove unicodes
                # r[i] = re.sub('[\'"]', '', r[i]) #remove annoying puncts
                r[i] = re.sub('[^\s\w\.\-\$_%\^&*#~+@"\']', ' ', r[i]) #remove annoying puncts
                for x in re.findall('(\.[a-z])', r[i]):
                    r[i] = re.sub('\.{0}'.format(x[1]), ' {0}'.format(x[1]), r[i])
                r[i] = re.sub('\s+', ' ', r[i])
                r[i] = r[i].strip()


class EntityTableDataExtraction(Extractor):
    def __init__(self) -> None:
        Extractor.__init__(self,
                           input_type=InputType.OBJECT,
                           category="data",
                           name="DigEntityTableDataExtractor")
        self.glossaries = dict()

    def add_glossary(self, glossary: List[str], attr_name: str) -> None:
        """
        Adds a glossary for the given attribute name
        :param glossary: a list of possible mentions of the attribute name
        :param attr_name: the attribute name (field name)
        """
        self.glossaries[attr_name] = glossary

    def wrap_value_with_context(self, value: dict, field_name: str, start: int=0, end: int=0) -> Extraction:
        """Wraps the final result"""
        return Extraction(value, self.name, start_token=start, end_token=end, tag=field_name)

    def extract(self, table: dict) -> List[Extraction]:
        """
        :param table: a table extracted by table extractor, as a json object
        :return: list of all extractions from the input table
        """
        if table['features']['max_cols_in_a_row'] != 2 and table['features']['no_of_rows'] < 2:
            return []
        results = list()
        for row in table['rows']:
            if len(row['cells']) != 2:
                continue
            text = [row['cells'][0]['text'], row['cells'][1]['text']]
            for field_name in self.glossaries.keys():
                if self.cell_matches_dict(text[0], self.glossaries[field_name]):
                    results.append(self.wrap_value_with_context(text[1], field_name))
                if self.cell_matches_dict(text[1], self.glossaries[field_name]):
                    results.append(self.wrap_value_with_context(text[0], field_name))
        return results

    def cell_matches_dict(self, cell_text: str, glossary: List[str]) -> bool:
        if any([self.cell_matches_text(cell_text, x) for x in glossary]):
            return True
        return False

    def cell_matches_text(self, cell_text: str, text: str) -> bool:
        cell_text = cell_text.lower()
        text = text.lower()
        if text in cell_text and float(len(cell_text))/float(len(text)) < 1.5:
            return True
        return False


class TableExtraction:
    @staticmethod
    def is_data_cell(cell):
        if cell.table:
            return False
        return True

    @staticmethod
    def is_data_row(row):
        if row.table:
            return False
        cell = row.findAll('th', recursive=False)
        cell.extend(row.findAll('td', recursive=False))
        for td in cell:
            if TableExtraction.is_data_cell(td) == False:
                return False
        return True

    @staticmethod
    def get_data_rows(table):
        data_rows = []
        rows = table.findAll('tr', recursive=False)
        if table.thead:
            rows.extend(table.thead.findAll('tr', recursive=False))
        if table.tbody:
            rows.extend(table.tbody.findAll('tr', recursive=False))
        for tr in rows:
            if TableExtraction.is_data_row(tr):
                data_rows.append(str(tr))
        return data_rows

    @staticmethod
    def is_data_table(table, k):
        rows = TableExtraction.get_data_rows(table)
        if len(rows) > k:
            return rows
        else:
            return False

    @staticmethod
    def mean(numbers):
        """ Computes mean of a list of numbers """
        return float(sum(numbers)) / max(len(numbers), 1)

    @staticmethod
    def _ss(data):
        """Return sum of square deviations of sequence data."""
        c = TableExtraction.mean(data)
        ss = sum((x-c)**2 for x in data)
        return ss

    @staticmethod
    def pstdev(data):
        """Calculates the population standard deviation."""
        n = len(data)
        if n < 2:
            return 0
            # raise ValueError('variance requires at least two data points')
        ss = TableExtraction._ss(data)
        pvar = ss/n # the population variance
        return pvar**0.5

    @staticmethod
    # Check if a string contains a digit
    def contains_digits(d):
        _digits = re.compile('\d')
        return bool(_digits.search(d))

    @staticmethod
    def gen_context(seq):
        seen = set()
        seen_add = seen.add
        uniq_list = [x for x in seq if not (x in seen or seen_add(x))]
        if len(uniq_list) > 5:
            uniq_list = uniq_list[:5]
        uniq_list = [x.replace("\t", "").replace("\r", "").replace("\n", "").strip() for x in uniq_list]
        if '' in uniq_list:
            uniq_list.remove('')
        if ' ' in uniq_list:
            uniq_list.remove(' ')
        return uniq_list

    def extract(self, html_doc, min_data_rows = 1):
        soup = BeautifulSoup(html_doc, 'html.parser')
        result_tables = list()
        tables = soup.findAll('table')
        for table in tables:
            tdcount = 0
            max_tdcount = 0
            img_count = 0
            href_count = 0
            inp_count = 0
            sel_count = 0
            colspan_count = 0
            colon_count = 0
            len_row = 0
            table_data = ""
            data_table = dict()
            row_list = list()
            rows = TableExtraction.is_data_table(table, min_data_rows)
            if rows != False:
                features = dict()
                row_len_list = list()
                avg_cell_len = 0
                avg_row_len_dev = 0
                for index_row, row in enumerate(rows):
                    row_dict = dict()
                    soup_row = BeautifulSoup(row, 'html.parser')
                    row_data = ' '.join(soup_row.stripped_strings)
                    row_data = row_data.replace("\\t", "").replace("\\r", "").replace("\\n", "")
                    if row_data != '':
                        row_len_list.append(len(row_data))
                        row_tdcount = len(soup_row.findAll('td')) + len(soup_row.findAll('th'))
                        if row_tdcount > max_tdcount:
                            max_tdcount = row_tdcount
                        tdcount += row_tdcount
                        img_count += len(soup_row.findAll('img'))
                        href_count += len(soup_row.findAll('a'))
                        inp_count += len(soup_row.findAll('input'))
                        sel_count += len(soup_row.findAll('select'))
                        colspan_count += row_data.count("colspan")
                        colon_count += row_data.count(":")
                        len_row += 1
                        table_data += row
                        # row_dict["row"] = str(row)
                        cell_list = list()
                        for index_col, td in enumerate(soup_row.findAll('th')):
                            cell_dict = dict()
                            cell_dict["cell"] = str(td)
                            # cell_dict["text"] = [{"result": {"value": ''.join(td.stripped_strings)}}]
                            cell_dict["text"] = ' '.join(td.stripped_strings)
                            cell_dict["id"] = 'row_{0}_col_{1}'.format(index_row, index_col)
                            avg_cell_len += len(cell_dict["text"])
                            cell_list.append(cell_dict)
                        for index_col, td in enumerate(soup_row.findAll('td')):
                            cell_dict = dict()
                            cell_dict["cell"] = str(td)
                            # cell_dict["text"] = [{"result": {"value": ''.join(td.stripped_strings)}}]
                            cell_dict["text"] = ' '.join(td.stripped_strings)
                            cell_dict["id"] = 'row_{0}_col_{1}'.format(index_row, index_col)
                            avg_cell_len += len(cell_dict["text"])
                            cell_list.append(cell_dict)
                        avg_row_len_dev += TableExtraction.pstdev([len(x["text"]) for x in cell_list])
                        row_dict["cells"] = cell_list
                        row_dict["text"] = self.row_to_text(cell_list)
                        row_dict["html"] = self.row_to_html(cell_list)
                        row_dict["id"] = "row_{}".format(index_row)
                        row_list.append(row_dict)

                # To avoid division by zero
                if len_row == 0:
                    tdcount = 1
                features["no_of_rows"] = len_row
                features["no_of_cells"] = tdcount
                features["max_cols_in_a_row"] = max_tdcount
                features["ratio_of_img_tags_to_cells"] = img_count*1.0/tdcount
                features["ratio_of_href_tags_to_cells"] = href_count*1.0/tdcount
                features["ratio_of_input_tags_to_cells"] = inp_count*1.0/tdcount
                features["ratio_of_select_tags_to_cells"] = sel_count*1.0/tdcount
                features["ratio_of_colspan_tags_to_cells"] = colspan_count*1.0/tdcount
                features["ratio_of_colons_to_cells"] = colon_count*1.0/tdcount
                features["avg_cell_len"] = avg_cell_len*1.0/tdcount
                features["avg_row_len"] = TableExtraction.mean(row_len_list)
                features["avg_row_len_dev"] = avg_row_len_dev*1.0/max(len_row, 1)

                avg_col_len = 0
                avg_col_len_dev = 0
                no_of_cols_containing_num = 0
                no_of_cols_empty = 0

                if colspan_count == 0.0 and \
                    len_row != 0 and \
                    (tdcount/(len_row * 1.0)) == max_tdcount:
                    col_data = dict()
                    for i in range(max_tdcount):
                        col_data['c_{0}'.format(i)] = []
                    soup_col = BeautifulSoup(table_data, 'html.parser')
                    for row in soup_col.findAll('tr'):
                        h_index = 0
                        h_bool = True
                        for col in row.findAll('th'):
                            col_content = ' '.join(col.stripped_strings)
                            h_bool = False
                            if col_content is None:
                                continue
                            else:
                                col_data['c_{0}'.format(h_index)].append(col_content)
                            h_index += 1
                        d_index = 0
                        if(h_index == 1 and h_bool == False):
                            d_index = 1
                        for col in row.findAll('td'):
                            col_content = ' '.join(col.stripped_strings)
                            if col_content is None:
                                d_index += 1
                                continue
                            else:
                                col_data['c_{0}'.format(d_index)].append(col_content)
                            d_index += 1

                    for key, value in col_data.items():
                        whole_col = ' '.join(value)
                        # avg_cell_len += float("%.2f" % mean([len(x) for x in value]))
                        avg_col_len += sum([len(x) for x in value])
                        avg_col_len_dev += TableExtraction.pstdev([len(x) for x in value])
                        no_of_cols_containing_num += 1 if TableExtraction.contains_digits(whole_col) is True else 0
                        # features["column_" + str(key) + "_is_only_num"] = whole_col.isdigit()
                        no_of_cols_empty += 1 if (whole_col == '') is True else 0
                # To avoid division by zero
                if max_tdcount == 0:
                    max_tdcount = 1
                features["avg_col_len"] = avg_col_len*1.0/max_tdcount
                features["avg_col_len_dev"] = avg_col_len_dev/max_tdcount
                features["no_of_cols_containing_num"] = no_of_cols_containing_num
                features["no_of_cols_empty"] = no_of_cols_empty
                data_table["features"] = features
                data_table["rows"] = row_list
                context_before = ' '.join(TableExtraction.gen_context(table.find_all_previous(string=True)))
                context_after = ' '.join(TableExtraction.gen_context(table.find_all_next(string=True)))
                table_rep = TableExtraction.gen_html(row_list)
                fingerprint = TableExtraction.create_fingerprint(table_rep)
                data_table["context_before"] = context_before
                data_table["context_after"] = context_after
                data_table["fingerprint"] = fingerprint
                data_table['html'] = table_rep
                data_table['text'] = self.table_to_text(row_list)
                result_tables.append(data_table)
                table.decompose()
        return dict(tables=result_tables, html_text=self.text_from_html(soup))

    @staticmethod
    def create_fingerprint(table):
        table = str(table)
        all_tokens = list(set(re.split('[^\w]+',table)))
        all_tokens = sorted(all_tokens)
        fingerprint = '-'.join(all_tokens)
        return fingerprint

    @staticmethod
    def row_to_html(cells):
        res = '<html><body><table>'
        for i, c in enumerate(cells):
            res += c['cell'] + '\n'
        res += '</table></body></html>'
        return res

    @staticmethod
    def row_to_text(cells):
        res = ''
        for i, c in enumerate(cells):
            res += c['text']
            if i < len(cells)-1:
                res += ' | '
        return res

    @staticmethod
    def table_to_text(rows):
        res = ''
        for row in rows:
            for i, c in enumerate(row['cells']):
                res += c['text']
                if i < len(row['cells']) - 1:
                    res += ' | '
            res += '\n'
        return res


    @staticmethod
    def gen_html(row_list):
        """ Return html table string from a list of data rows """
        table = "<table>"
        for row in row_list:
            table += "<tr>"
            cells = row["cells"]
            for cell in cells:
                table += str(cell["cell"])
            table += "</tr>"
        table += "</table>"
        return table

    @staticmethod
    def remove_tables(html_doc, min_data_rows = 1):
        soup = BeautifulSoup(html_doc, 'html.parser')
        tables = soup.findAll('table')
        for table in tables:
            rows = TableExtraction.is_data_table(table, min_data_rows)
            if rows != False:
                table.decompose()

        return soup

    def tag_visible(self, element):
        if element.parent.name in ['style', 'script', 'head', 'title', 'meta', '[document]']:
            return False
        if isinstance(element, Comment):
            return False
        return True

    def text_from_html(self, soup):
        texts = soup.findAll(text=True)
        visible_texts = filter(self.tag_visible, texts)
        # print([x.strip() for x in visible_texts])
        # exit(0)
        return u" ".join(t.strip() for t in visible_texts if t.strip() != "")


class TableExtractor(Extractor):
    """
    **Description**
        Adding explanation here

    Examples:
        ::

            table_extractor = TableExtractor()
            table_extractor.extract(html=html_str,
                                    return_text=True)

    """

    tableExtractorInstance = TableExtraction()

    def __init__(self) -> None:
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="content",
                           name="DigTableExtractor")

    def _wrap_value_with_context(self, value: dict or str, field_name: str, start: int=0, end: int=0) -> Extraction:
        """Wraps the final result"""
        return Extraction(value, self.name, start_token=start, end_token=end, tag=field_name)

    def extract(self, html: str, return_text: bool=False) -> List[Extraction]:
        """
        Args:
            html (str): raw html of the page
            return_text (bool): if True, return the visible text in the page
                                removing all the data tables

        Returns:
            List[Extraction]: a list of Extractions

        """
        results = list()
        temp_res = TableExtractor.tableExtractorInstance.extract(html)
        if return_text:
            results.append(self._wrap_value_with_context(temp_res['html_text'], "text_without_tables"))
        results.extend(map(lambda t: self._wrap_value_with_context(t, "tables"), temp_res['tables']))
        return results



import sys
import os
import importlib
from argparse import ArgumentParser, RawDescriptionHelpFormatter

# module name should NOT starts with '__' (double underscore)
# module name can not be in 'help', '--help', 'h', '-h'.
dir_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cli')
handlers = list(filter(lambda x: not x.startswith('__'),
                       [os.path.splitext(fname)[0] for fname in os.listdir(dir_path)]))


def help_info():
    print('Usage:')
    print('\t', 'python -m etk <command> [options]')
    print('Available commands:')
    print('\t', ', '.join(handlers))
    print('For the help information of a particular command:')
    print('\t', 'python -m etk <command> -h')
    exit()

if __name__ == '__main__':
    """
    Usage:
        python -m etk <command> [options]
    Example:
        python -m etk dummy --test "this is a test"
    """
    if len(sys.argv) <= 1:
        print('No command\n')
        help_info()

    cmd = sys.argv[1]
    sub_cmd = sys.argv[2:] if len(sys.argv) >= 3 else []

    if cmd in ('help', '--help', 'h', '-h'):
        help_info()

    if cmd not in handlers:
        print('Unknown command\n')
        help_info()

    # load module
    sys.path.append(dir_path)
    mod = importlib.import_module(cmd)

    # parse arguments
    parser = ArgumentParser(prog='python -m etk {}'.format(cmd), formatter_class=RawDescriptionHelpFormatter)
    mod.add_arguments(parser)
    args = parser.parse_args(args=sub_cmd)

    # run
    mod.run(args)



from typing import List
from enum import Enum, auto
from bs4 import BeautifulSoup
from bs4.element import Comment
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from etk.extractors.readability.readability import Document


class Strategy(Enum):
    """
    ALL_TEXT: return all visible text in an HTML page
    MAIN_CONTENT_STRICT: MAIN_CONTENT_STRICT: return the main content of the page without boiler plate (menu, ads...)
    MAIN_CONTENT_RELAXED: variant of MAIN_CONTENT_STRICT with less strict rules
    """
    ALL_TEXT = auto()
    MAIN_CONTENT_STRICT = auto()
    MAIN_CONTENT_RELAXED = auto()


class HTMLContentExtractor(Extractor):
    """
    **Description**
        This class extracts text from HTML pages. Uses readability and BeautifulSoup.

    Examples:
        ::

            html_content_extractor = HTMLContentExtractor()
            html_content_extractor.extract(text=input_doc,
                                        strategy=Strategy.ALL_TEXT)

    """

    def __init__(self):
        Extractor.__init__(self,
                           input_type=InputType.HTML,
                           category="HTML extractor",
                           name="HTML content extractor")

    def extract(self, html_text: str, strategy: Strategy=Strategy.ALL_TEXT) \
            -> List[Extraction]:
        """
        Extracts text from an HTML page using a variety of strategies

        Args:
            html_text (str): html page in string
            strategy (enum[Strategy.ALL_TEXT, Strategy.MAIN_CONTENT_RELAXED, Strategy.MAIN_CONTENT_STRICT]): one of
            Strategy.ALL_TEXT, Strategy.MAIN_CONTENT_STRICT and Strategy.MAIN_CONTENT_RELAXED

        Returns:
             List[Extraction]: typically a singleton list with the extracted text
        """

        if html_text:
            if strategy == Strategy.ALL_TEXT:
                soup = BeautifulSoup(html_text, 'html.parser')
                texts = soup.findAll(text=True)
                visible_texts = filter(self._tag_visible, texts)
                all_text = u" ".join(t.strip() for t in visible_texts)
                return [Extraction(all_text, self.name)]
            else:
                relax = strategy == Strategy.MAIN_CONTENT_RELAXED
                readable = Document(html_text, recallPriority=relax).summary(html_partial=False)
                clean_text = BeautifulSoup(readable.encode('utf-8'), 'lxml').strings
                readability_text = ' '.join(clean_text)
                return [Extraction(readability_text, self.name)]
        else:
            return []

    @staticmethod
    def _tag_visible(element):
        if element.parent.name in ['style', 'script', 'head', 'title', 'meta', '[document]']:
            return False
        if isinstance(element, Comment):
            return False
        return True




from typing import List
from etk.extraction import Extraction
from etk.extractor import Extractor, InputType

from enum import Enum, auto
from langid import classify
from langdetect import detect


class LanguageIdentificationExtractor(Extractor):
    """
    **Description**
        Identify the language used in text, returning the identifier language using ISO 639-1 codes

        Uses two libraries:
        - https://github.com/davidjurgens/equilid
        - https://github.com/saffsd/langid.py

        TODO: define Enum to select which method to use.
        TODO: define dictionary to translate ISO 639-3 to ISO 639-1 codes
        https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes, perhaps there is an online source that has this

    Examples:
        ::

            language_identification_extractor = LanguageIdentificationExtractor()
            language_identification_extractor.extract(text=input_stri,
                                                    method=IdentificationTool.LANGDETECT.name)

    """

    def __init__(self):
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="Text extractor",
                           name="Language Identification")

    def extract(self, text: str, method: str) -> List[Extraction]:
        """

        Args:
            text (str): any text, can contain HTML
            method (Enum[IdentificationTool.LANGID, IdentificationTool.LANGDETECT]): specifies which of the two
            algorithms to use

        Returns:
            List(Extraction): an extraction containing the language code used in the text. Returns the empty list of
            the extractor fails to identify the language in the text.

        """
        if method == IdentificationTool.LANGID.name:
            language = classify(text)[0]
            return [Extraction(value=language, extractor_name=self.name)]

        elif method == IdentificationTool.LANGDETECT.name:
            try:
                language = detect(text)
            except:
                language = 'unknown'

            if language == 'unknown':
                return list()
            else:
                return [Extraction(value=language, extractor_name=self.name)]

        else:
            return list()


class IdentificationTool(Enum):
    LANGID = auto()
    LANGDETECT = auto()



import pyexcel_io
import pyexcel_xlsx
import os
import csv
import datetime
from etk.document import Document
from typing import List
from io import StringIO
from etk.etk_exceptions import InvalidArgumentsError, InvalidFilePathError


class CsvProcessor(object):
    """
                        heading_row: int,
                        content_start_row: int,
                        heading_columns: (int, int),
                        content_end_row: int,
                        blank_row_ends_content: bool,
                        remove_leading_trailing_whitespace: bool,
                        required_columns: list[str]
                        column_name_prefix: str
    """

    def __init__(self, etk, **mapping_spec) -> None:
        self.etk = etk

        self.heading_row = mapping_spec.get("heading_row")
        if self.heading_row is not None:
            self.heading_row = self.heading_row - 1

        if self.heading_row is not None:
            self.content_start_row = mapping_spec.get("content_start_row", self.heading_row + 2) - 1

        if self.heading_row is None:
            self.content_start_row = mapping_spec.get("content_start_row", 1) - 1

        # how about if heading_col is not present? read until first empty cell?
        self.heading_columns = mapping_spec.get("heading_columns")

        if self.heading_columns is not None:
            self.heading_columns = (self.heading_columns[0] - 1, self.heading_columns[1])

        # if not present, default read until an empty row
        self.content_end_row = mapping_spec.get("content_end_row")

        # if set to false, read until EOF
        self.blank_row_ends_content = mapping_spec.get("ends_with_blank_row", True)

        # remove all white space of the value
        self.remove_leading_trailing_whitespace = \
            mapping_spec.get("remove_leading_trailing_whitespace", True)

        # remove all empty rows before the content
        self.remove_leading_empty_rows = \
            mapping_spec.get("remove_leading_empty_rows", True)

        self.required_columns = mapping_spec.get("required_columns")

        if mapping_spec.get("column_name_prefix"):
            self.column_name_prefix = mapping_spec.get("column_name_prefix")
        else:
            self.column_name_prefix = "C"

        self._get_data_function = {
            ".csv": pyexcel_io.get_data,
            ".tsv": pyexcel_io.get_data,
            ".xls": pyexcel_xlsx.get_data,
            ".xlsx": pyexcel_xlsx.get_data
        }

    def tabular_extractor(self, table_str: str = None, filename: str = None,
                          file_content=None,
                          file_type=None,
                          sheet_name: str = None,
                          dataset: str = None,
                          nested_key: str = None,
                          doc_id_field: str = None,
                          encoding=None) -> List[Document]:
        """
        Read the input file/content and return a list of Document(s)
        Args:
            table_str: use this parameter, if you are 100% sure that the content is a csv
            filename: use this parameter if the file extension is one of tab, csv, tsv, xls, xlsx
            file_content: if the input has some arbitrary extension, read it yourself and pass the contents along
            file_type: use this parameter with file_content, can be tsv, csv, etc
            sheet_name: sheet name as in xls or xlsx files
            dataset: user provided string to be added to output Document(s)
            nested_key: user provided string to be added to output Document(s)
            doc_id_field: specify this field(should be present in the input file), its value will be used as doc_id

        Returns: List[Document]

        """
        data = list()

        if table_str is not None and filename is not None:
            raise InvalidArgumentsError(message="for arguments 'table_str' and 'filename', please specify only one "
                                                "argument!")

        elif table_str is not None:
            f = StringIO(table_str)
            reader = csv.reader(f, delimiter=',')
            for row in reader:
                data.append(row)
        elif filename is not None:
            # always read the entire file first
            fn, extension = os.path.splitext(filename)
            extension = extension.lower()

            if extension in self._get_data_function:
                get_data = self._get_data_function[extension]
            else:
                # in pyexcel we trust
                # if there is an extension we have not mapped, just let pyexcel figure it out
                get_data = pyexcel_io.get_data

            try:
                if file_content and file_type:
                    data = get_data(file_content, file_type=file_type, auto_detect_datetime=False,
                                    auto_detect_float=False, encoding=encoding if encoding else "utf-8")
                else:
                    data = get_data(filename, auto_detect_datetime=False,
                                    auto_detect_float=False, encoding=encoding if encoding else "utf-8")
            except:
                try:
                    data = get_data(filename, auto_detect_datetime=False,
                                    auto_detect_float=False, encoding="latin_1")
                except:
                    data = get_data(filename, auto_detect_datetime=False,
                                    auto_detect_float=False, encoding="utf-8-sig")

            if extension == '.xls' or extension == '.xlsx':
                if sheet_name is None:
                    sheet_name = list(data.keys())[0]

                data = data[sheet_name]
            else:
                data = data[file_type] if file_type else data[fn.split('/')[-1] + extension]

        table_content, heading = self.content_recognizer(data)

        return self.create_documents(rows=table_content,
                                     heading=heading,
                                     file_name=filename,
                                     dataset=dataset,
                                     nested_key=nested_key,
                                     doc_id_field=doc_id_field)

    def content_recognizer(self, data: List[List[str]]) -> tuple((List[List[str]], List[str])):
        heading = list()
        # process heading
        if self.heading_row is not None:
            heading, col_start, col_end = self.process_heading(data[self.heading_row])
            # if heading_row is specified, discards/overwrite the heading_columns
            self.heading_columns = (col_start, col_end)
        else:
            heading = None

        # handle row first:
        if self.content_end_row is None:
            data = data[self.content_start_row:]

        if self.content_end_row is not None:
            data = data[self.content_start_row:self.content_end_row]

        data, col_start, col_end, row_count = self.process_by_row(data)
        if self.heading_columns is None:
            self.heading_columns = (col_start, col_end)

        # handle col:
        data = self.extract_row_content(data)

        return data, heading

    @staticmethod
    def process_heading(heading: List[str]) -> tuple((List[str], int, int)):
        processed_heading = list()
        col_start, col_end = 0, len(heading)
        for i in range(len(heading)):
            if not heading[i] and col_start == 0:
                continue
            elif heading[i] and col_start == 0:
                col_start = i
                processed_heading.append(heading[i])
            elif not heading[i] and col_start != 0:
                col_end = i
                break
            else:
                processed_heading.append(heading[i])

        return processed_heading, col_start - 1, col_end

    # slicing table by start and end col
    def extract_row_content(self, sheet: List[List[str]]) -> List[List[str]]:
        valid_row = list()
        for row in sheet:
            valid_row.append(row[self.heading_columns[0]:self.heading_columns[1]])

        return valid_row

    # remove all leading empty rows and figure out the start and end col
    def process_by_row(self, sheet: List[List[str]]) -> tuple((List[List[str]], int, int, int)):
        col_min = float("inf")
        col_max = 0
        row_count = 0
        valid_row = list()

        for row in sheet:
            # first row & row is empty & remove all leading empty rows
            if row_count == 0 and not any(row) and self.remove_leading_empty_rows:
                continue
            elif row_count == 0 and not any(row) and not self.remove_leading_empty_rows:
                row_count += 1
                continue
            # not first row & row is empty & end with blank row
            elif row_count != 0 and not any(row) and self.blank_row_ends_content:
                break
            elif row_count != 0 and not any(row) and not self.blank_row_ends_content:
                row_count += 1
                continue

            #  find first non-empty col
            idx = list(map(bool, row)).index(True)
            col_min = min(col_min, idx)
            col_max = max(len(row), col_max)
            row_count += 1
            valid_row.append(row)

        return valid_row, col_min, col_max, row_count

    def create_documents(self, rows: List[List[str]],
                         heading: List[str] = None,
                         file_name: str = None,
                         dataset: str = None,
                         nested_key: str = None,
                         doc_id_field: str = None) -> List[Document]:
        documents = list()
        # etk = ETK()
        if self.heading_row is None and self.required_columns is not None:
            raise InvalidArgumentsError("cannot match the required columns since heading is not specified")

        # get the heading line index of required columns
        list_idx = list()
        if self.required_columns is not None:
            for i in range(len(heading)):
                if heading[i] in self.required_columns:
                    list_idx.append(i)
        # filter each row
        for row in rows:
            # if the row is empty, skip it
            if not any(row):
                continue
            # if some required field is missing, skip it
            is_required_not_empty = all(row[i] for i in list_idx)
            if not is_required_not_empty:
                continue
            # convert datetime obj to ISO format str
            row = self.datetime_to_string(row)
            # create doc for each row
            doc = dict()
            for i in range(0, self.heading_columns[1] - self.heading_columns[0]):
                if heading is not None:
                    key = heading[i]
                else:
                    key = self.column_name_prefix + str(i)

                if i >= len(row):
                    doc[key] = ''
                else:
                    doc[key] = row[i]

            cdr_doc = dict()
            if nested_key is not None:
                cdr_doc[nested_key] = doc
            else:
                cdr_doc = doc

            if file_name is not None:
                cdr_doc['file_name'] = file_name
            if dataset is not None:
                cdr_doc['dataset'] = dataset

            doc_id = None
            if doc_id_field:
                doc_id = cdr_doc.get(doc_id_field, None)
            documents.append(self.etk.create_document(cdr_doc, doc_id=doc_id))

        return documents

    # convert datetime object of list to ISO format string
    @staticmethod
    def datetime_to_string(row: List[object]) -> List[object]:
        for i in range(len(row)):
            if type(row[i]) is datetime.date:
                row[i] = row[i].isoformat()

        return row



from lxml.html import tostring
import logging
import lxml.html
import re, sys

from .cleaners import normalize_spaces, clean_attributes
from .encoding import get_encoding

utf8_parser = lxml.html.HTMLParser(encoding='utf-8')

def build_doc(page):
    if isinstance(page, str):
        enc = None
        page_unicode = page
    else:
        enc = get_encoding(page) or 'utf-8'
        page_unicode = page.decode(enc, 'replace')
    doc = lxml.html.document_fromstring(page_unicode.encode('utf-8', 'replace'), parser=utf8_parser)
    return doc, enc

def js_re(src, pattern, flags, repl):
    return re.compile(pattern, flags).sub(src, repl.replace('$', '\\'))


def normalize_entities(cur_title):
    entities = {
        u'\u2014':'-',
        u'\u2013':'-',
        u'&mdash;': '-',
        u'&ndash;': '-',
        u'\u00A0': ' ',
        u'\u00AB': '"',
        u'\u00BB': '"',
        u'&quot;': '"',
    }
    for c, r in entities.items():
        if c in cur_title:
            cur_title = cur_title.replace(c, r)

    return cur_title

def norm_title(title):
    return normalize_entities(normalize_spaces(title))

def get_title(doc):
    title = doc.find('.//title')
    if title is None or title.text is None or len(title.text) == 0:
        return '[no-title]'

    return norm_title(title.text)

def add_match(collection, text, orig):
    text = norm_title(text)
    if len(text.split()) >= 2 and len(text) >= 15:
        if text.replace('"', '') in orig.replace('"', ''):
            collection.add(text)

def shorten_title(doc):
    title = doc.find('.//title')
    if title is None or title.text is None or len(title.text) == 0:
        return ''

    title = orig = norm_title(title.text)

    candidates = set()

    for item in ['.//h1', './/h2', './/h3']:
        for e in list(doc.iterfind(item)):
            if e.text:
                add_match(candidates, e.text, orig)
            if e.text_content():
                add_match(candidates, e.text_content(), orig)

    for item in ['#title', '#head', '#heading', '.pageTitle', '.news_title', '.title', '.head', '.heading', '.contentheading', '.small_header_red']:
        for e in doc.cssselect(item):
            if e.text:
                add_match(candidates, e.text, orig)
            if e.text_content():
                add_match(candidates, e.text_content(), orig)

    if candidates:
        title = sorted(candidates, key=len)[-1]
    else:
        for delimiter in [' | ', ' - ', ' :: ', ' / ']:
            if delimiter in title:
                parts = orig.split(delimiter)
                if len(parts[0].split()) >= 4:
                    title = parts[0]
                    break
                elif len(parts[-1].split()) >= 4:
                    title = parts[-1]
                    break
        else:
            if ': ' in title:
                parts = orig.split(': ')
                if len(parts[-1].split()) >= 4:
                    title = parts[-1]
                else:
                    title = orig.split(': ', 1)[1]

    if not 15 < len(title) < 150:
        return orig

    return title

def get_body(doc):
    [ elem.drop_tree() for elem in doc.xpath('.//script | .//link | .//style') ]
    raw_html = tostring(doc.body or doc).decode("utf-8")
    print(raw_html)
    cleaned = clean_attributes(raw_html)
    try:
        #BeautifulSoup(cleaned) #FIXME do we really need to try loading it?
        return cleaned
    except Exception: #FIXME find the equivalent lxml error
        #logging.error("cleansing broke html content: %s\n---------\n%s" % (raw_html, cleaned))
        return raw_html



from etk.ontology_api import Ontology
from etk.ontology_report_generator import OntologyReportGenerator


def add_arguments(parser):
    parser.description = 'Generate HTML report for the input ontology files'
    parser.add_argument('files', nargs='+', help='Input turtle files.')
    parser.add_argument('--no-validation', action='store_false', dest='validation', default=True,
                        help='Don\'t perform domain and range validation.')
    parser.add_argument('-o', '--output', dest='out', default='ontology-doc.html',
                        help='Location of generated HTML report.')
    parser.add_argument('-i', '--include-undefined-classes', action='store_true',
                        dest='include_class', default=False, help='Include those undefined classes '
                                                                  'but referenced by others.')
    parser.add_argument('-t', '--include-turtle', action='store_true', dest='include_turtle',
                        default=False, help='Include turtle related to this entity. NOTE: this may '
                                            'takes longer time.')
    parser.add_argument('-q', '--quiet', action='store_true', dest='quiet', default=False,
                        help='Suppress warning.')
    parser.add_argument('--exclude-warning', action='store_true', dest='exclude_warning',
                        default=False, help='Exclude warning messages in HTML report')
    parser.add_argument('--list-auxiliary', action='store_true', dest='list_auxiliary',
                        default=False, help='Show auxiliary line for list')


def run(args):
    contents = [open(f).read() for f in args.files]
    ontology = Ontology(contents, validation=args.validation, include_undefined_class=args.include_class,
                        quiet=args.quiet)
    doc_content = OntologyReportGenerator(ontology).generate_html_report(include_turtle=args.include_turtle,
                                                                         exclude_warning=args.exclude_warning,
                                                                         list_auxiliary_line=args.list_auxiliary)

    with open(args.out, "w") as f:
        f.write(doc_content)




import unittest
from etk.extractors.decoding_value_extractor import DecodingValueExtractor


class TestDictionaryExtractor(unittest.TestCase):
    def test_dictionary_extractor(self) -> None:
        decoding_dict = {
            'CA': 'California',
            'ny': 'New York',
            'AZ': ' Arizona',
            ' TX ': 'Texas',
            ' fl': 'Florida',
        }
        values = ['ca', 'CA', ' CA', ' ca', 'NY', ' ny', 'Az', 'AZ', 'az ', 'tx', 'tx ', 'TX', 'fl', 'FL', 'fl ']

        de_default = DecodingValueExtractor(decoding_dict, 'default_decoding') # strip_key and not case_sensitive
        de_case_sensitive = DecodingValueExtractor(decoding_dict, 'default_decoding', case_sensitive=True)
        de_not_strip_key = DecodingValueExtractor(decoding_dict, 'default_decoding', strip_key=False)
        de_strip_value = DecodingValueExtractor(decoding_dict, 'default_decoding', strip_value=True)

        results = list()
        results.append([de_default.extract(v)[0].value for v in values if de_default.extract(v)])
        results.append([de_case_sensitive.extract(v)[0].value for v in values if de_case_sensitive.extract(v)])
        results.append([de_not_strip_key.extract(v)[0].value for v in values if de_not_strip_key.extract(v)])
        results.append([de_strip_value.extract(v)[0].value for v in values if de_strip_value.extract(v)])

        expected = [
            [
                'California', 'California', 'California', 'California', 'New York', 'New York', ' Arizona',
                ' Arizona', ' Arizona', 'Texas', 'Texas', 'Texas', 'Florida', 'Florida', 'Florida'
            ],
            [
                'California', 'California', 'New York', ' Arizona', 'Texas', 'Florida', 'Florida'
            ],
            [
                'California', 'California', 'New York', ' Arizona', ' Arizona'
            ],
            [
                'California', 'California', 'California', 'California', 'New York', 'New York', 'Arizona',
                'Arizona', 'Arizona', 'Texas', 'Texas', 'Texas', 'Florida', 'Florida', 'Florida'
            ],
        ]

        self.assertEqual(results[:-1], expected[:-1])


if __name__ == '__main__':
    unittest.main()


import unittest
from etk.extractors.spacy_rule_extractor import SpacyRuleExtractor
import spacy
from etk.extractors.html_metadata_extractor import HTMLMetadataExtractor
import json


class TestSpacyRuleExtractor(unittest.TestCase):

    def test_SpacyRuleExtractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/news.html', 'r') as f:
            sample_html = f.read()

        sample_rules = json.load(open('etk/unit_tests/ground_truth/sample_spacy_rule.json'))

        title_extraction = hme.extract(sample_html, extract_title=True)[0].value

        sample_rule_extractor = SpacyRuleExtractor(spacy.load("en_core_web_sm"), sample_rules, "dummy")
        extractions = sample_rule_extractor.extract(title_extraction)
        expected_extraction = 'Trump'
        self.assertEqual(extractions[0].value, expected_extraction)


if __name__ == '__main__':
    unittest.main()



import warnings
import sys
import argparse

from etk.extractors.html_metadata_extractor import HTMLMetadataExtractor


def add_arguments(parser):
    """
       Parse arguments
       Args:
           parser (argparse.ArgumentParser)
       """
    parser.description = 'Examples:\n' \
                         'python -m etk regex_extractor pattern /tmp/date.txt\n' \
                         'cat /tmp/date.txt | python -m etk regex_extractor pattern'
    parser.add_argument('pattern', nargs='?', type=str, default=sys.stdin)
    parser.add_argument('input_file', nargs='?', type=argparse.FileType('r'), default=sys.stdin)


def run(args):
    """
    Args:
        args (argparse.Namespace)
    """
    html_metadata_extractor = HTMLMetadataExtractor()

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        extractions = html_metadata_extractor.extract(html_text=args.input_file)
        for e in extractions:
            print(e.value)



import unittest, json

from etk.document_selector import DefaultDocumentSelector
from etk.etk import ETK
from etk.knowledge_graph_schema import KGSchema

sample_input = {
    "dataset": "unittestUnittestUNITTEST",
    "website": "abcABCAbc",
    "url": "zxcZXCZxc",
    "projects": [
        {
            "name": "etk",
            "description": "version 2 of etk, implemented by Runqi, Dongyu, Sylvia, Amandeep and others."
        },
        {
            "name": "rltk",
            "description": "record linkage toolkit, implemented by Pedro, Mayank, Yixiang and several students."
        }
    ]
}

kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))
etk = ETK(kg_schema=kg_schema)

class TestDocumentSelector(unittest.TestCase):

    def test_datasets_condition(self) -> None:
        doc = etk.create_document(sample_input)
        default_doc_selector = DefaultDocumentSelector()
        res_true = default_doc_selector.select_document(doc, datasets=[".*unittest", ".*abc"])
        res_false = default_doc_selector.select_document(doc, datasets=[".*abc", ".*hhhh"])
        self.assertEqual(True, res_true)
        self.assertEqual(False, res_false)

    def test_url_patterns_condition(self) -> None:
        doc = etk.create_document(sample_input)
        default_doc_selector = DefaultDocumentSelector()
        res_true = default_doc_selector.select_document(doc, url_patterns=[".*unittest", ".*zxc"])
        res_false = default_doc_selector.select_document(doc, url_patterns=[".*ZXc", ".*hhhh"])
        self.assertEqual(True, res_true)
        self.assertEqual(False, res_false)

    def test_website_patterns_condition(self) -> None:
        doc = etk.create_document(sample_input)
        default_doc_selector = DefaultDocumentSelector()
        res_true = default_doc_selector.select_document(doc, website_patterns=[".*unittest", ".*abc"])
        res_false = default_doc_selector.select_document(doc, website_patterns=[".*ABc", ".*hhhh"])
        self.assertEqual(True, res_true)
        self.assertEqual(False, res_false)

    def test_json_paths_and_json_paths_regex(self) -> None:
        doc = etk.create_document(sample_input)
        default_doc_selector = DefaultDocumentSelector()
        res_true = default_doc_selector.select_document(doc, json_paths=["$.website"],
                                                        json_paths_regex=[".*unittest", ".*abc"])
        res_false = default_doc_selector.select_document(doc, json_paths=["$.website"], json_paths_regex=[".*hhhh"])
        self.assertEqual(True, res_true)
        self.assertEqual(False, res_false)

    def test_all_condition(self) -> None:
        doc = etk.create_document(sample_input)
        default_doc_selector = DefaultDocumentSelector()
        res_true = default_doc_selector.select_document(doc, datasets=[".*unittest", ".*abc"],
                                                        url_patterns=[".*unittest", ".*zxc"],
                                                        website_patterns=[".*unittest", ".*abc"],
                                                        json_paths=["$.website"],
                                                        json_paths_regex=[".*unittest", ".*abc"])
        res_false = default_doc_selector.select_document(doc, datasets=[".*abc", ".*hhhh"],
                                                         url_patterns=[".*ZXc", ".*hhhh"],
                                                         website_patterns=[".*ABc", ".*hhhh"], json_paths=["$.website"],
                                                         json_paths_regex=[".*hhhh"])
        self.assertEqual(True, res_true)
        self.assertEqual(False, res_false)


if __name__ == '__main__':
    unittest.main()



from SPARQLWrapper import SPARQLWrapper, JSON
from collections import OrderedDict
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from typing import List
import requests


class DBpediaSpotlightExtractor(Extractor):
    """
    **Description**
        This extractor takes a string of text as input, uses DBPedia API to annotate words and phrases in the text input.

    Examples:
        ::

            dbpedia_spotlight_extractor = DBpediaSpotlightExtractor(search_url='http://model.dbpedia-spotlight.org/en/annotate',
                                                                    get_attr=False,
                                                                    get_attr_url="http://dbpedia.org/sparql")
            dbpedia_spotlight_extractor.extract(text=input_doc,
                                                filter=['Person', 'Place', 'Organisation'])

    """
    def __init__(self, extractor_name: str, search_url: str, get_attr=False,
                 get_attr_url="http://dbpedia.org/sparql"):
        Extractor.__init__(self, input_type=InputType.TEXT,
                           category="built_in_extractor",
                           name=extractor_name)
        self._search_url = search_url
        self._get_attr = get_attr
        self._get_attr_url = get_attr_url

    def extract(self, text: str, confidence=0.5, filter=['Person', 'Place', 'Organisation']) -> List[Extraction]:
        """
            Extract with the input text, confidence and fields filter to be used.
            Args:
                text (str): text input to be annotated
                confidence (float): the confidence of the annotation
                filter (List[str]): the fields that to be extracted

            Returns:
                List[Extraction]
        """

        filter = ','.join(filter)
        search_data = [('confidence', confidence),
                       ('text', text),
                       ('types', filter)]
        search_headers = {'Accept': 'application/json'}
        r = requests.post(self._search_url,
                          data=search_data,
                          headers=search_headers)
        results = r.json()
        last_results = self._combiner(results)
        return last_results

    def _combiner(self, results: dict) -> List[Extraction]:
        return_result = list()
        if "Resources" in results:
            resources_results = results["Resources"]
            for one_result in resources_results:
                types = one_result['@types'].split(',')
                values = {'surface_form': one_result['@surfaceForm'],
                          'uri': one_result['@URI'],
                          'types': types,
                          'similarity_scores': float(one_result['@similarityScore'])}
                if self._get_attr:
                    attr = self._attr_finder(one_result['@URI'])
                    values['attributes'] = attr
                return_result.append(Extraction(confidence=float(results['@confidence']),
                                                extractor_name=self.name,
                                                start_char=int(one_result['@offset']),
                                                end_char=int(one_result['@offset']) + len(
                                                    one_result['@surfaceForm']),
                                                value=values))

            return return_result
        return list()

    def _attr_finder(self, uri) -> dict:
        sparql = SPARQLWrapper(self._get_attr_url)
        sparql.setQuery("SELECT distinct * WHERE {<" + uri + "> ?link ?resource}")
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()

        attr = OrderedDict()
        cnt_attr = 0
        for one_item in results['results']['bindings']:
            if ('xml:lang' in one_item['resource']) and (one_item['resource']['xml:lang'] != 'en'):
                pass
            else:
                attr_key = one_item['link']['value'].split('/')[-1]
                attr_val = one_item['resource']['value']
                if attr_key not in attr:
                    if cnt_attr < 100:
                        attr[attr_key] = [attr_val]
                        cnt_attr += 1
                else:
                    attr[attr_key].append(attr_val)
        return attr



import datetime
import hashlib
import json
from typing import Dict
import uuid


class Utility(object):

    @staticmethod
    def make_json_serializable(doc: Dict):
        """
        Make the document JSON serializable. This is a poor man's implementation that handles dates and nothing else.
        This method modifies the given document in place.

        Args:
            doc: A Python Dictionary, typically a CDR object.

        Returns: None

        """
        for k, v in doc.items():
            if isinstance(v, datetime.date):
                doc[k] = v.strftime("%Y-%m-%d")
            elif isinstance(v, datetime.datetime):
                doc[k] = v.isoformat()

    @staticmethod
    def create_doc_id_from_json(doc) -> str:
        """
        Docs with identical contents get the same ID.
        Args:
            doc:

        Returns: a string with the hash of the given document.

        """
        return hashlib.sha256(json.dumps(doc, sort_keys=True).encode('utf-8')).hexdigest()

    @staticmethod
    def create_doc_id_string(any_string):
        """
        Creates sha256 has of a string
        :param any_string: input string
        :return: sha256 hash of any_string
        """
        try:
            return hashlib.sha256(any_string).hexdigest()
        except:
            # probably failed because of unicode
            return hashlib.sha256(any_string.encode('utf-8')).hexdigest()

    @staticmethod
    def create_uuid():
        return str(uuid.uuid4())

    @staticmethod
    def create_description_from_json(doc_json):
        description = ''
        for key in doc_json:
            description += '"' + key + '":"' + str(doc_json[key]) + '", <br/>'
        description += '}'
        return description



from typing import List, Dict
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from etk.tokenizer import Tokenizer
from spacy.matcher import Matcher
from spacy import attrs
from spacy.tokens import span, doc
from etk.extractors.util.util import tf_transfer
import copy
import itertools
import sys
import re

FLAG_DICT = {
    20: attrs.FLAG20,
    21: attrs.FLAG21,
    22: attrs.FLAG22,
    23: attrs.FLAG23,
    24: attrs.FLAG24,
    25: attrs.FLAG25,
    26: attrs.FLAG26,
    27: attrs.FLAG27,
    28: attrs.FLAG28,
    29: attrs.FLAG29,
    30: attrs.FLAG30,
    31: attrs.FLAG31,
    32: attrs.FLAG32,
    33: attrs.FLAG33,
    34: attrs.FLAG34,
    35: attrs.FLAG35,
    36: attrs.FLAG36,
    37: attrs.FLAG37,
    38: attrs.FLAG38,
    39: attrs.FLAG39,
    40: attrs.FLAG40,
    41: attrs.FLAG41,
    42: attrs.FLAG42,
    43: attrs.FLAG43,
    44: attrs.FLAG44,
    45: attrs.FLAG45,
    46: attrs.FLAG46,
    47: attrs.FLAG47,
    48: attrs.FLAG48,
    49: attrs.FLAG49,
    50: attrs.FLAG50,
    51: attrs.FLAG51,
    52: attrs.FLAG52,
    53: attrs.FLAG53,
    54: attrs.FLAG54,
    55: attrs.FLAG55,
    56: attrs.FLAG56,
    57: attrs.FLAG57,
    58: attrs.FLAG58,
    59: attrs.FLAG59,
    60: attrs.FLAG60,
    61: attrs.FLAG61,
    62: attrs.FLAG62,
    63: attrs.FLAG63
}

POS_MAP = {
    "AUX": "AUX",
    "EOL": "EOL",
    "CCONJ": "CCONJ",
    "SCONJ": "SCONJ",
    "noun": "NOUN",
    "pronoun": "PROPN",
    "proper noun": "PROPN",
    "determiner": "DET",
    "symbol": "SYM",
    "adjective": "ADJ",
    "conjunction": "CONJ",
    "verb": "VERB",
    "pre/post-position": "ADP",
    "adverb": "ADV",
    "particle": "PART",
    "interjection": "INTJ",
    "X": "X",
    "NUM": "NUM",
    "SPACE": "SPACE",
    "PRON": "PRON"
}


FLAG_ID = 20


class SpacyRuleExtractor(Extractor):
    """
    **Description**
        This extractor takes a spaCy rule as reference and extracts the substring
        which matches the given spaCy rule.

    Examples:
        ::

            rules = json.load(open('path_to_spacy_rules.json', "r"))
            sample_rules = rules["test_SpacyRuleExtractor_word_1"]
            spacy_rule_extractor = SpacyRuleExtractor(nlp=nlp,
                                                     rules=sample_rules)
            spacy_rule_extractor.extract(text=text)

    """

    def __init__(self,
                 nlp,
                 rules: Dict,
                 extractor_name: str) -> None:
        """
        Initialize the extractor, storing the rule information and construct spacy rules
        Args:
            nlp
            rules (Dict): spacy rules
            extractor_name: str

        Returns:
        """

        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="spacy_rule_extractor",
                           name=extractor_name)
        self._rules = rules["rules"]
        self._nlp = copy.deepcopy(nlp)
        self._tokenizer = Tokenizer(self._nlp)
        self._matcher = Matcher(self._nlp.vocab)
        self._field_name = rules["field_name"] if "field_name" in rules else extractor_name
        self._rule_lst = {}
        self._hash_map = {}
        for idx, a_rule in enumerate(self._rules):
            this_rule = Rule(a_rule, self._nlp)
            self._rule_lst[this_rule.identifier + "rule_id##" + str(idx)] = this_rule

    def extract(self, text: str) -> List[Extraction]:
        """
        Extract from text

        Args:
            text (str): input str to be extracted.

        Returns:
            List[Extraction]: the list of extraction or the empty list if there are no matches.
        """

        doc = self._tokenizer.tokenize_to_spacy_doc(text)
        self._load_matcher()

        matches = [x for x in self._matcher(doc) if x[1] != x[2]]
        pos_filtered_matches = []
        neg_filtered_matches = []
        for idx, start, end in matches:
            span_doc = self._tokenizer.tokenize_to_spacy_doc(doc[start:end].text)
            this_spacy_rule = self._matcher.get(idx)
            relations = self._find_relation(span_doc, this_spacy_rule)
            rule_id, _ = self._hash_map[idx]
            this_rule = self._rule_lst[rule_id]
            if self._filter_match(doc[start:end], relations, this_rule.patterns):
                value = self._form_output(doc[start:end], this_rule.output_format, relations, this_rule.patterns)
                if this_rule.polarity:
                    pos_filtered_matches.append((start, end, value, rule_id, relations))
                else:
                    neg_filtered_matches.append((start, end, value, rule_id, relations))

        return_lst = []
        if pos_filtered_matches:
            longest_lst_pos = self._get_longest(pos_filtered_matches)
            if neg_filtered_matches:
                longest_lst_neg = self._get_longest(neg_filtered_matches)
                return_lst = self._reject_neg(longest_lst_pos, longest_lst_neg)
            else:
                return_lst = longest_lst_pos

        extractions = []
        for (start, end, value, rule_id, relation) in return_lst:
            this_extraction = Extraction(value=value,
                                         extractor_name=self.name,
                                         start_token=start,
                                         end_token=end,
                                         start_char=doc[start].idx,
                                         end_char=doc[end-1].idx+len(doc[end-1]),
                                         rule_id=rule_id.split("rule_id##")[0],
                                         match_mapping=relation)
            extractions.append(this_extraction)

        return extractions

    def _load_matcher(self) -> None:
        """
        Add constructed spacy rule to Matcher
        """
        for id_key in self._rule_lst:
            if self._rule_lst[id_key].active:
                pattern_lst = [a_pattern.spacy_token_lst for a_pattern in self._rule_lst[id_key].patterns]

                for spacy_rule_id, spacy_rule in enumerate(itertools.product(*pattern_lst)):
                    self._matcher.add(self._construct_key(id_key, spacy_rule_id), None, list(spacy_rule))

    def _filter_match(self, span: span, relations: Dict, patterns: List) -> bool:
        """
        Filter the match result according to prefix, suffix, min, max ...
        Args:
            span: span
            relations: Dict
            patterns: List of pattern

        Returns: bool
        """

        for pattern_id, a_pattern in enumerate(patterns):
            token_range = relations[pattern_id]
            if token_range:
                tokens = [x for x in span[token_range[0]:token_range[1]]]
                if a_pattern.type == "word":
                    if not self._pre_suf_fix_filter(tokens, a_pattern.prefix, a_pattern.suffix):
                        return False
                if a_pattern.type == "shape":
                    if not (self._full_shape_filter(tokens, a_pattern.full_shape)
                            and self._pre_suf_fix_filter(tokens, a_pattern.prefix,a_pattern.suffix)):
                        return False
                if a_pattern.type == "number":
                    if not self._min_max_filter(tokens, a_pattern.min, a_pattern.max):
                        return False
        return True

    @staticmethod
    def _get_longest(value_lst: List) -> List:
        """
        Get the longest match for overlap
        Args:
            value_lst: List

        Returns: List
        """

        value_lst.sort()
        result = []
        pivot = value_lst[0]
        start, end = pivot[0], pivot[1]
        pivot_e = end
        pivot_s = start
        for idx, (s, e, v, rule_id, _) in enumerate(value_lst):
            if s == pivot_s and pivot_e < e:
                pivot_e = e
                pivot = value_lst[idx]
            elif s != pivot_s and pivot_e < e:
                result.append(pivot)
                pivot = value_lst[idx]
                pivot_e = e
                pivot_s = s
        result.append(pivot)
        return result

    @staticmethod
    def _reject_neg(pos_lst: List, neg_lst: List) -> List:
        """
        Reject some positive matches according to negative matches
        Args:
            pos_lst: List
            neg_lst: List

        Returns: List
        """

        pos_lst.sort()
        neg_lst.sort()
        result = []
        pivot_pos = pos_lst[0]
        pivot_neg = neg_lst[0]
        while pos_lst:
            if pivot_pos[1] <= pivot_neg[0]:
                result.append(pivot_pos)
                pos_lst.pop(0)
                if pos_lst:
                    pivot_pos = pos_lst[0]
            elif pivot_pos[0] >= pivot_neg[1]:
                neg_lst.pop(0)
                if not neg_lst:
                    result += pos_lst
                    break
                else:
                    pivot_neg = neg_lst[0]
            else:
                pos_lst.pop(0)
                if pos_lst:
                    pivot_pos = pos_lst[0]
        return result

    @staticmethod
    def _pre_suf_fix_filter(t: List, prefix: str, suffix: str) -> bool:
        """
        Prefix and Suffix filter
        Args:
            t: List, list of tokens
            prefix: str
            suffix: str

        Returns: bool
        """

        if prefix:
            for a_token in t:
                if a_token._.n_prefix(len(prefix)) != prefix:
                    return False
        if suffix:
            for a_token in t:
                if a_token._.n_suffix(len(suffix)) != suffix:
                    return False

        return True

    @staticmethod
    def _min_max_filter(t: List, min_v: str, max_v: str) -> bool:
        """
        Min and Max filter
        Args:
            t: List, list of tokens
            min_v: str
            max_v: str

        Returns: bool
        """

        def tofloat(value):
            try:
                float(value)
                return float(value)
            except ValueError:
                return False

        for a_token in t:
            if not tofloat(a_token.text):
                return False
            else:
                if min_v and tofloat(min_v):
                    this_v = tofloat(a_token.text)
                    if this_v < tofloat(min_v):
                        return False
                if max_v and tofloat(max_v):
                    this_v = tofloat(a_token.text)
                    if this_v > tofloat(max_v):
                        return False

        return True

    @staticmethod
    def _full_shape_filter(t: List, shapes: List) -> bool:
        """
        Shape filter
        Args:
            t: List, list of tokens
            shapes: List

        Returns: bool
        """

        if shapes:
            for a_token in t:
                if a_token._.full_shape not in shapes:
                    return False

        return True

    @staticmethod
    def _form_output(span_doc: span, output_format: str, relations: Dict, patterns: List) -> str:
        """
        Form an output value according to user input of output_format
        Args:
            span_doc: span
            format: str
            relations: Dict
            patterns: List

        Returns: str
        """

        format_value = []
        output_inf = [a_pattern.in_output for a_pattern in patterns]
        for i in range(len(output_inf)):
            token_range = relations[i]
            if token_range and output_inf[i]:
                format_value.append(span_doc[token_range[0]:token_range[1]].text)

        if not output_format:
            return " ".join(format_value)

        result_str = re.sub("{}", " ".join(format_value), output_format)

        positions = re.findall("{[0-9]+}", result_str)

        if not positions:
            return result_str

        position_indices = [int(x[1:-1]) for x in positions]
        if max(position_indices) < len(format_value):
            result_str = result_str.format(*format_value)
        else:
            try:
                result_str = result_str.format("", *format_value)
            except:
                positions = [x for x in positions if int(x[1:-1]) > len(format_value)-1 or int(x[1:-1]) < 0]
                for pos in positions:
                    result_str = result_str.replace(pos, "")
                result_str = result_str.format(*format_value)

        return result_str

    def _construct_key(self, rule_id: str, spacy_rule_id:int) -> int:
        """
        Use a mapping to store the information about rule_id for each matches, create the mapping key here
        Args:
            rule_id: str
            spacy_rule_id:int

        Returns: int
        """

        hash_key = (rule_id, spacy_rule_id)
        hash_v = hash(hash_key) + sys.maxsize + 1
        self._hash_map[hash_v] = hash_key
        return hash_v

    def _find_relation(self, span_doc: doc, r: List) -> Dict:
        """
        Get the relations between the each pattern in the spacy rule and the matches
        Args:
            span_doc: doc
            r: List

        Returns: Dict
        """

        rule = r[1][0]
        span_pivot = 0
        relation = {}
        for e_id, element in enumerate(rule):
            if not span_doc[span_pivot:]:
                for extra_id, _, in enumerate(rule[e_id:]):
                    relation[e_id+extra_id] = None
                break
            new_doc = self._tokenizer.tokenize_to_spacy_doc(span_doc[span_pivot:].text)
            if "OP" not in element:
                relation[e_id] = (span_pivot, span_pivot+1)
                span_pivot += 1
            else:
                if e_id < len(rule)-1:
                    tmp_rule_1 = [rule[e_id]]
                    tmp_rule_2 = [rule[e_id+1]]
                    tmp_matcher = Matcher(self._nlp.vocab)
                    tmp_matcher.add(0, None, tmp_rule_1)
                    tmp_matcher.add(1, None, tmp_rule_2)
                    tmp_matches = sorted([x for x in tmp_matcher(new_doc) if x[1] != x[2]], key=lambda a: a[1])

                    if not tmp_matches:
                        relation[e_id] = None
                    else:
                        matches_1 = [x for x in tmp_matches if x[0] == 0 and x[1] == 0]
                        if not matches_1:
                            relation[e_id] = None
                        else:
                            _, s1, e1 = matches_1[0]
                            matches_2 = [x for x in tmp_matches if x[0] == 1]
                            if not matches_2:
                                relation[e_id] = (span_pivot, span_pivot + e1)
                                span_pivot += e1
                            else:
                                _, s2, e2 = matches_2[0]
                                if e1 <= s2:
                                    relation[e_id] = (span_pivot, span_pivot + e1)
                                    span_pivot += e1
                                else:
                                    relation[e_id] = (span_pivot, span_pivot + s2)
                                    span_pivot += s2
                else:
                    relation[e_id] = (span_pivot, len(span_doc))

        return relation


class Pattern(object):
    """
    class pattern represent each token

    For each token, we let user specify constrains for tokens. Some attributes are spacy build-in attributes,
    which can be used with rule-based matching: https://spacy.io/usage/linguistic-features#section-rule-based-matching
    Some are custom attributes, need to apply further filtering after we get matches
    """

    def __init__(self, d: Dict, nlp) -> None:
        """
        Initialize a pattern, construct spacy token for matching according to type
        Args:
            d: Dict
            nlp

        Returns:
        """

        self.type = d["type"]
        self.in_output = tf_transfer(d["is_in_output"])
        self.max = d["maximum"]
        self.min = d["minimum"]
        self.prefix = d["prefix"]
        self.suffix = d["suffix"]
        self.full_shape = d.get("shapes")

        if self.type == "word":
            self.spacy_token_lst = self._construct_word_token(d, nlp)
        elif self.type == "shape":
            self.spacy_token_lst = self._construct_shape_token(d)
        elif self.type == "number":
            self.spacy_token_lst = self._construct_number_token(d, nlp)
        elif self.type == "punctuation":
            self.spacy_token_lst = self._construct_punctuation_token(d, nlp)
        elif self.type == "linebreak":
            self.spacy_token_lst = self._construct_linebreak_token(d)

    def _construct_word_token(self, d: Dict, nlp) -> List[Dict]:
        """
        Construct a word token
        Args:
            d: Dict
            nlp

        Returns: List[Dict]
        """

        result = []
        if len(d["token"]) == 1:
            if tf_transfer(d["match_all_forms"]):
                this_token = {attrs.LEMMA: nlp(d["token"][0])[0].lemma_}
            else:
                this_token = {attrs.LOWER: d["token"][0].lower()}
            result.append(this_token)
            if d["capitalization"]:
                result = self._add_capitalization_constrain(result, d["capitalization"], d["token"])

        elif not d["token"]:
            if tf_transfer(d["contain_digit"]):
                this_token = {attrs.IS_ASCII: True, attrs.IS_PUNCT: False}
            else:
                this_token = {attrs.IS_ALPHA: True}
            if tf_transfer(d["is_out_of_vocabulary"]) and not tf_transfer(d["is_in_vocabulary"]):
                this_token[attrs.IS_OOV] = True
            elif not tf_transfer(d["is_out_of_vocabulary"]) and tf_transfer(d["is_in_vocabulary"]):
                this_token[attrs.IS_OOV] = False
            result.append(this_token)
            if d["length"]:
                result = self._add_length_constrain(result, d["length"])
            if d["capitalization"]:
                result = self._add_capitalization_constrain(result, d["capitalization"], d["token"])

        else:
            if "match_all_forms" in d and not tf_transfer(d["match_all_forms"]):
                global FLAG_ID
                token_set = set(d["token"])

                def is_selected_token(x):
                    return x in token_set

                FLAG_DICT[FLAG_ID] = nlp.vocab.add_flag(is_selected_token)
                this_token = {FLAG_DICT[FLAG_ID]: True}
                FLAG_ID += 1
                result.append(this_token)

            else:
                token_set = [nlp(x)[0].lemma_ for x in set(d["token"])]
                for a_lemma in token_set:
                    this_token = {attrs.LEMMA: a_lemma}
                    result.append(this_token)

            if d["capitalization"]:
                result = self._add_capitalization_constrain(result, d["capitalization"], d["token"])

        result = self._add_common_constrain(result, d)
        if d["part_of_speech"]:
            result = self._add_pos_constrain(result, d["part_of_speech"])

        return result

    def _construct_shape_token(self, d: Dict) -> List[Dict]:
        """
        Construct a shape token
        Args:
            d: Dict

        Returns: List[Dict]
        """

        result = []
        if not d["shapes"]:
            this_token = {attrs.IS_ASCII: True}
            result.append(this_token)
        else:
            for shape in d["shapes"]:
                this_shape = self._generate_shape(shape)
                this_token = {attrs.SHAPE: this_shape}
                result.append(copy.deepcopy(this_token))

        result = self._add_common_constrain(result, d)
        if d["part_of_speech"]:
            result = self._add_pos_constrain(result, d["part_of_speech"])

        return result

    def _construct_number_token(self, d: Dict, nlp) -> List[Dict]:
        """
        Construct a shape token
        Args:
            d: Dict
            nlp

        Returns: List[Dict]
        """

        result = []
        if not d["numbers"]:
            this_token = {attrs.LIKE_NUM: True}
            result.append(this_token)
            if d["length"]:
                result = self._add_length_constrain(result, d["length"])
        elif len(d["numbers"]) == 1:
            this_token = {attrs.ORTH: str(d["numbers"][0])}
            result.append(this_token)
        else:
            global FLAG_ID
            number_set = set(d["numbers"])

            def is_selected_number(x):
                return x in number_set

            FLAG_DICT[FLAG_ID] = nlp.vocab.add_flag(is_selected_number)
            this_token = {FLAG_DICT[FLAG_ID]: True}
            FLAG_ID += 1
            result.append(this_token)
        result = self._add_common_constrain(result, d)
        return result

    def _construct_punctuation_token(self, d: Dict, nlp) -> List[Dict]:
        """
        Construct a shape token
        Args:
            d: Dict
            nlp

        Returns: List[Dict]
        """

        result = []
        if not d["token"]:
            this_token = {attrs.IS_PUNCT: True}
        elif len(d["token"]) == 1:
            this_token = {attrs.ORTH: d["token"][0]}
        else:
            global FLAG_ID
            punct_set = set(d["token"])

            def is_selected_punct(x):
                return x in punct_set

            FLAG_DICT[FLAG_ID] = nlp.vocab.add_flag(is_selected_punct)
            this_token = {FLAG_DICT[FLAG_ID]: True}
            FLAG_ID += 1
        result.append(this_token)
        result = self._add_common_constrain(result, d)
        return result

    def _construct_linebreak_token(self, d: Dict) -> List[Dict]:
        """
        Construct a shape token
        Args:
            d: Dict

        Returns: List[Dict]
        """

        result = []
        num_break = int(d["length"][0]) if d["length"] else 1
        if num_break:
            s = ''
            for i in range(num_break):
                s += '\n'
            this_token = {attrs.LOWER: s}
            result.append(this_token)
            s += ' '
            this_token = {attrs.LOWER: s}
            result.append(this_token)
        result = self._add_common_constrain(result, d)

        return result

    @staticmethod
    def _add_common_constrain(token_lst: List[Dict], d: Dict) -> List[Dict]:
        """
        Add common constrain for every token type, like "is_required"
        Args:
            token_lst: List[Dict]
            d: Dict

        Returns: List[Dict]
        """

        result = []
        for a_token in token_lst:
            if not tf_transfer(d["is_required"]):
                a_token["OP"] = "?"
            result.append(a_token)
        return result

    @staticmethod
    def _add_length_constrain(token_lst: List[Dict], lengths: List) -> List[Dict]:
        """
        Add length constrain for some token type, create cross production
        Args:
            token_lst: List[Dict]
            lengths: List

        Returns: List[Dict]
        """

        result = []
        for a_token in token_lst:
            for length in lengths:
                if type(length) == str and length and length.isdigit():
                    a_token[attrs.LENGTH] = int(length)
                    result.append(copy.deepcopy(a_token))
                elif type(length) == int:
                    a_token[attrs.LENGTH] = int(length)
                    result.append(copy.deepcopy(a_token))
        return result

    @staticmethod
    def _add_pos_constrain(token_lst: List[Dict], pos_tags: List) -> List[Dict]:
        """
        Add pos tag constrain for some token type, create cross production
        Args:
            token_lst: List[Dict]
            pos_tags: List

        Returns: List[Dict]
        """

        result = []
        for a_token in token_lst:
            for pos in pos_tags:
                a_token[attrs.POS] = POS_MAP[pos]
                result.append(copy.deepcopy(a_token))
        return result

    @staticmethod
    def _add_capitalization_constrain(token_lst: List[Dict], capi_lst: List, word_lst: List) -> List[Dict]:
        """
        Add capitalization constrain for some token type, create cross production
        Args:
            token_lst: List[Dict]
            capi_lst: List
            word_lst: List

        Returns: List[Dict]
        """

        result = []
        for a_token in token_lst:
            if "exact" in capi_lst and word_lst != []:
                for word in word_lst:
                    token = copy.deepcopy(a_token)
                    token[attrs.ORTH] = word
                    result.append(token)
            if "lower" in capi_lst:
                token = copy.deepcopy(a_token)
                token[attrs.IS_LOWER] = True
                result.append(token)
            if "upper" in capi_lst:
                token = copy.deepcopy(a_token)
                token[attrs.IS_UPPER] = True
                result.append(token)
            if "title" in capi_lst:
                token = copy.deepcopy(a_token)
                token[attrs.IS_TITLE] = True
                result.append(token)
            if "mixed" in capi_lst:
                token = copy.deepcopy(a_token)
                token[attrs.IS_UPPER] = False
                token[attrs.IS_LOWER] = False
                token[attrs.IS_TITLE] = False
                result.append(token)
        return result

    @staticmethod
    def _generate_shape(word: str) -> str:
        """
        Recreate shape from a token input by user
        Args:
            word: str

        Returns: str
        """

        def counting_stars(w) -> List[int]:
            count = [1]
            for i in range(1, len(w)):
                if w[i - 1] == w[i]:
                    count[-1] += 1
                else:
                    count.append(1)

            return count

        shape = ""
        p = 0
        for c in counting_stars(word):
            if c > 4:
                shape += word[p:p + 4]
            else:
                shape += word[p:p + c]
            p = p + c

        return shape


class Rule(object):
    """
        Class Rule represent each matching rule, each rule contains many pattern
    """

    def __init__(self, d: Dict, nlp) -> None:
        """
        Storing information for each Rule, create list of Pattern for a rule
        Args:
            d: Dict
            nlp

        Returns:
        """

        self.dependencies = d["dependencies"] if "dependencies" in d else []
        self.description = d["description"] if "description" in d else ""
        self.active = tf_transfer(d["is_active"])
        self.identifier = d["identifier"]
        self.output_format = d["output_format"]
        self.polarity = tf_transfer(d["polarity"])
        self.patterns = []
        for pattern_idx, a_pattern in enumerate(d["pattern"]):
            this_pattern = Pattern(a_pattern, nlp)
            self.patterns.append(this_pattern)




import warnings
import sys
import argparse

from etk.extractors.date_extractor import DateExtractor

date_extractor = DateExtractor()


def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    parser.description = 'Examples:\n' \
                         'python -m etk date_extractor /tmp/date.txt\n' \
                         'cat /tmp/date.txt | python -m etk date_extractor'
    parser.add_argument('input_file', nargs='?', type=argparse.FileType('r'), default=sys.stdin)


def run(args):
    """
    Args:
        args (argparse.Namespace)
    """
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        for line in args.input_file:
            extractions = date_extractor.extract(line)
            for e in extractions:
                print(e.value)



from typing import List, Dict
from bs4 import BeautifulSoup
from etk.extraction_provenance_record import ExtractionProvenanceRecord
from etk.extraction import Extractable, Extraction
from etk.extractor import Extractor, InputType
from etk.segment import Segment
from etk.tokenizer import Tokenizer
from etk.knowledge_graph import KnowledgeGraph
from etk.utilities import Utility
from etk.etk_exceptions import ErrorPolicy, ExtractorValueError
import warnings


class Document(Segment):
    """
        This class wraps raw CDR documents and provides a convenient API for ETK
        to query elements of the document and to update the document with the results
        of extractors.
        """

    def __init__(self, etk, cdr_document: Dict, mime_type, url, doc_id=None) -> None:

        """
        Wrapper object for CDR documents.

        Args:
            etk (ETK): embed the etk object so that docs have access to global info.
            cdr_document (JSON): the raw CDR document received in ETK.

        Returns: the wrapped CDR document

        """
        Segment.__init__(self, json_path="$", _value=cdr_document, _document=self)
        self.etk = etk
        self.cdr_document = cdr_document
        self.mime_type = mime_type
        self.url = url
        if doc_id:
            self.cdr_document["doc_id"] = doc_id
        self.extraction_provenance_records = list()
        if self.etk.kg_schema:
            self.kg = KnowledgeGraph(self.etk.kg_schema, self.etk.ontology, self)
        else:
            self.kg = None
            if not self.etk.kg_schema:
                self.etk.log("Schema not found.", "warning", self.doc_id, self.url)
        self._provenance_id_index = 0
        self._provenances = dict()
        self._jsonpath_provenances = dict()
        self._kg_provenances = dict()

    @property
    def provenance_id_index(self) -> int:
        return self._provenance_id_index

    @property
    def provenances(self) -> Dict:
        return self._provenances

    @property
    def jsonpath_provenances(self) -> Dict:
        return self._jsonpath_provenances

    @property
    def kg_provenances(self) -> Dict:
        return self._kg_provenances

    def provenance_id_index_incrementer(self):
        self._provenance_id_index += 1

    @property
    def document(self):
        """
        Still thinking about this, having the parent doc inside each extractable is convenient to avoid
        passing around the parent doc to all methods.

        Returns: the parent Document
        """
        return self

    def select_segments(self, jsonpath: str) -> List[Segment]:
        """
        Dereferences the json_path inside the document and returns the selected elements.
        This method should compile and cache the compiled json_path in case the same path
        is reused by multiple extractors.

        Args:
            jsonpath (str): a valid JSON path.

        Returns: A list of Segments object that contains the elements selected by the json path.
        """
        path = self.etk.parse_json_path(jsonpath)
        matches = path.find(self.cdr_document)

        segments = list()
        for a_match in matches:
            this_segment = Segment(str(a_match.full_path), a_match.value, self)
            segments.append(this_segment)

        return segments

    def extract(self, extractor: Extractor, extractable: Extractable = None, tokenizer: Tokenizer = None,
                joiner: str = "  ", **options) -> List[Extraction]:

        """
        Invoke the extractor on the given extractable, accumulating all the extractions in a list.

        Args:
            extractor (Extractor):
            extractable (extractable):
            tokenizer: user can pass custom tokenizer if extractor wants token
            joiner: user can pass joiner if extractor wants text
            options: user can pass arguments as a dict to the extract() function of different extractors

        Returns: List of Extraction, containing all the extractions.

        """
        if not extractable:
            extractable = self

        if not tokenizer:
            tokenizer = self.etk.default_tokenizer

        extracted_results = list()

        if extractor.input_type == InputType.TOKENS:
            if self.etk.error_policy == ErrorPolicy.PROCESS:
                if isinstance(extractable.value, list):
                    self.etk.log(
                        "Extractor needs tokens, tokenizer needs string to tokenize, got list, converting to string",
                        "warning", self.doc_id, self.url)
                    warnings.warn(
                        "Extractor needs tokens, tokenizer needs string to tokenize, got list, converting to string")
                elif isinstance(extractable.value, dict):
                    self.etk.log(
                        "Extractor needs tokens, tokenizer needs string to tokenize, got dict, converting to string",
                        "warning", self.doc_id, self.url)
                    warnings.warn(
                        "Extractor needs tokens, tokenizer needs string to tokenize, got dict, converting to string")
                tokens = extractable.get_tokens(tokenizer)
                if tokens:
                    extracted_results = extractor.extract(tokens, **options)
            else:
                raise ExtractorValueError(
                    "Extractor needs string, tokenizer needs string to tokenize, got " + str(type(extractable.value)))

        elif extractor.input_type == InputType.TEXT:
            if self.etk.error_policy == ErrorPolicy.PROCESS:
                if isinstance(extractable.value, list):
                    self.etk.log("Extractor needs string, got extractable value as list, converting to string",
                                 "warning", self.doc_id, self.url)
                    warnings.warn("Extractor needs string, got extractable value as list, converting to string")
                elif isinstance(extractable.value, dict):
                    self.etk.log("Extractor needs string, got extractable value as dict, converting to string",
                                 "warning", self.doc_id, self.url)
                    warnings.warn("Extractor needs string, got extractable value as dict, converting to string")
                text = extractable.get_string(joiner)
                if text:
                    extracted_results = extractor.extract(text, **options)
            else:
                # raise ExtractorValueError("Extractor needs string, got " + str(type(extractable.value)))
                # TODO: Yixiang - needs to be handled properly
                pass

        elif extractor.input_type == InputType.OBJECT:
            extracted_results = extractor.extract(extractable.value, **options)

        elif extractor.input_type == InputType.HTML:
            if bool(BeautifulSoup(extractable.value, "html.parser").find()):
                extracted_results = extractor.extract(extractable.value, **options)
            else:
                # raise ExtractorValueError("Extractor needs HTML, got non HTML string")
                # TODO: Yixiang - needs to be handled properly
                pass

        try:
            jsonPath = extractable.full_path
        except AttributeError:
            jsonPath = None

        for e in extracted_results:
            # for the purpose of provenance hierarrchy tracking, a parent's id for next generation.
            e.prov_id = self.provenance_id_index
            extraction_provenance_record: ExtractionProvenanceRecord = ExtractionProvenanceRecord(
                e.prov_id, jsonPath, e.provenance["extractor_name"],
                e.provenance["start_char"], e.provenance["end_char"], e.provenance["confidence"], self,
                extractable.prov_id)
            self._provenances[e.prov_id] = extraction_provenance_record

            # for the purpose of provenance hierarchy tracking
            self.provenance_id_index_incrementer()
            self.create_provenance(extraction_provenance_record)

        return extracted_results

    @property
    def doc_id(self):
        """
        Returns: the doc_id of the CDR document

        """
        return self.cdr_document.get("doc_id")

    @doc_id.setter
    def doc_id(self, new_doc_id):
        """

        Args:
           new_doc_id ():

        Returns:

        """
        self.cdr_document["doc_id"] = new_doc_id

    def create_provenance(self, extraction_provenance_record: ExtractionProvenanceRecord) -> None:
        if "provenances" not in self.cdr_document:
            self.cdr_document["provenances"] = []
        self.cdr_document["provenances"].append(self.get_dict_extraction_provenance(extraction_provenance_record))

    def get_dict_extraction_provenance(self, extraction_provenance_record: ExtractionProvenanceRecord):
        prov_dict = dict()
        prov_dict["@id"] = extraction_provenance_record.id
        prov_dict["@type"] = "extraction_provenance_record"
        prov_dict["method"] = extraction_provenance_record.method
        prov_dict["confidence"] = extraction_provenance_record.extraction_confidence
        if extraction_provenance_record.origin_record.full_path is not None:
            jsonpath = extraction_provenance_record.origin_record.full_path
            if jsonpath in self._jsonpath_provenances:
                self._jsonpath_provenances[jsonpath].append(extraction_provenance_record.id)
            else:
                self._jsonpath_provenances[jsonpath] = [extraction_provenance_record.id]
            origin_dict = dict()
            origin_dict["path"] = jsonpath
            origin_dict["start_char"] = extraction_provenance_record.origin_record.start_char
            origin_dict["end_char"] = extraction_provenance_record.origin_record.end_char
            prov_dict["origin_record"] = origin_dict
        if extraction_provenance_record.parent_extraction_provenance is not None:
            prov_dict["parent_provenance_id"] = extraction_provenance_record.parent_extraction_provenance
        return prov_dict

    def insert_kg_into_cdr(self):
        if self.kg and self.kg.value:
            self.cdr_document["knowledge_graph"] = self.kg.value

    def build_knowledge_graph(self, json_ontology: dict) -> List:
        """
        The idea if to be able to build a json knowledge graph from a json like ontology representation, eg:
         kg_object_ontology = {
            "uri": doc.doc_id,
            "country": doc.select_segments("$.Location"),
            "type": [
                "Event",
                doc.extract(self.incomp_decoder, doc.select_segments("$.Incomp")[0]),
                doc.extract(self.int_decoder, doc.select_segments("$.Int")[0])
            ]
        }
        Currently json ontology representation is supported, might add new ways later
        Args:
            json_ontology: a json ontology representation of a knowledge graph

        Returns: returns  a list of nested documents created

        """
        nested_docs = list()
        if json_ontology:
            for key in list(json_ontology):
                j_values = json_ontology[key]
                if not isinstance(j_values, list):
                    j_values = [j_values]
                for j_value in j_values:
                    if not isinstance(j_value, dict):
                        if self.kg:
                            if key not in ['doc_id', 'uri']:
                                self.kg.add_value(key, value=j_value)
                    else:
                        """Now we have to create a nested document, assign it a doc_id and 
                           add the doc_id to parent document's knowledge graph"""
                        child_doc_id = None
                        if 'uri' in j_value:
                            child_doc_id = j_value['uri']
                        elif 'doc_id' in j_value:
                            child_doc_id = j_value['doc_id']

                        child_doc = Document(self.etk, cdr_document=dict(), mime_type='json', url='')
                        nested_docs.extend(child_doc.build_knowledge_graph(j_value))

                        if not child_doc_id:
                            child_doc_id = Utility.create_doc_id_from_json(child_doc.kg._kg)

                        if self.kg:
                            self.kg.add_value(key, value=child_doc_id)
                        child_doc.cdr_document["doc_id"] = child_doc_id

                        nested_docs.append(child_doc)

        return nested_docs

    def add_type(self, type_):
        self.kg.add_value('@type', value=type_)

    def with_type(self, type_):
        if type_:
            self.add_type(type_)
        return self



import unittest
from etk.extractors.table_extractor import TableExtractor, EntityTableDataExtraction


class TestTableExtractor(unittest.TestCase):
    test_doc = '''
        <html>
        <body>
        test random text!
        <table><tbody>
        <tr> 
        <th>caliber:</th><td>0.45 mm</td>
        </tr>
        <tr> 
        <th>manufacturer:</th><td>WXC</td>
        </tr>
        <tr> 
        <th>country:</th><td>Russia</td>
        </tr>
        <tr> 
        <th>price:</th><td>400 eur</td>
        </tr>
        </tbody></table>
        <body>
        </html>
    '''

    def test_table_extractor(self) -> None:
        my_table_extractor = TableExtractor()
        res = my_table_extractor.extract(TestTableExtractor.test_doc)
        self.assertEqual(len(res), 1)
        res = my_table_extractor.extract(TestTableExtractor.test_doc, True)
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].value, "test random text!")

    def test_entity_table_data_extractor(self) -> None:
        expected_res = [("caliber", "0.45 mm"),
                        ("manufacturer", "WXC"),
                        ("location", "Russia"),
                        ("price", "400 eur")]
        my_table_extractor = TableExtractor()
        table_data_extractor = EntityTableDataExtraction()
        table_data_extractor.add_glossary(["caliber", "calibre"], "caliber")
        table_data_extractor.add_glossary(["manufacturer"], "manufacturer")
        table_data_extractor.add_glossary(["price"], "price")
        table_data_extractor.add_glossary(["country"], "location")
        res = my_table_extractor.extract(TestTableExtractor.test_doc)
        res = table_data_extractor.extract(res[0].value)
        res = [(x.tag, x.value) for x in res]
        self.assertEqual(res, expected_res)


if __name__ == '__main__':
    unittest.main()


import spacy
import re
from spacy.tokenizer import Tokenizer as spacyTokenizer
from spacy.tokens import Token, Doc
from typing import List


class Tokenizer(object):
    """
    Abstract class used for all tokenizer implementations.
    """

    def __init__(self, nlp=spacy.load('en_core_web_sm'), keep_multi_space=True) -> None:
        """Load vocab, more vocab are available at: https://spacy.io/models/en"""
        self.nlp = nlp
        self._keep_multi_space = keep_multi_space

        """Custom tokenizer"""
        self.nlp.tokenizer = self.custom_tokenizer()

    @property
    def keep_multi_space(self):
        """
        Returns: the doc_id of the CDR document

        """
        return self._keep_multi_space

    @keep_multi_space.setter
    def keep_multi_space(self, new_keep_multi_space):
        """

        Args:
           new_keep_multi_space:

        Returns:

        """
        self._keep_multi_space = new_keep_multi_space

    def tokenize(self, text: str, customize=True, disable=[]) -> List[Token]:
        """
        Tokenize the given text, returning a list of tokens. Type token: class spacy.tokens.Token

        Args:
            text (string):

        Returns: [tokens]

        """

        """Tokenize text"""
        if not self.keep_multi_space:
            text = re.sub(' +', ' ', text)
        # disable spacy parsing, tagging etc as it takes a long time if the text is short
        tokens = self.nlp(text, disable=disable)
        if customize:
            tokens = [self.custom_token(a_token) for a_token in tokens]

        return tokens

    def tokenize_to_spacy_doc(self, text: str) -> Doc:
        """
        Tokenize the given text, returning a spacy doc. Used for spacy rule extractor

        Args:
            text (string):

        Returns: Doc

        """
        if not self.keep_multi_space:
            text = re.sub(' +', ' ', text)
        doc = self.nlp(text, disable=['parser'])
        for a_token in doc:
            self.custom_token(a_token)

        return doc

    def custom_tokenizer(self) -> spacyTokenizer:
        """
        Custom tokenizer
        For future improvement, look at https://spacy.io/api/tokenizer, https://github.com/explosion/spaCy/issues/1494
        """
        prefix_re = re.compile(r'''^[\[()\-.,@#$%^&*?|<~+_:;>!"']''')
        infix_re = re.compile(r'''[\[()\-,@#$%^&*?|<~+_:;>!"']|(?<![0-9])\.|\.(?![0-9])|\n+ ''')
        return spacyTokenizer(self.nlp.vocab, rules=None, prefix_search=prefix_re.search, suffix_search=None,
                              infix_finditer=infix_re.finditer, token_match=None)

    @staticmethod
    def custom_token(spacy_token) -> Token:
        """
        Function for token attributes extension, methods extension
        Use set_extension method.
        Reference: https://spacy.io/api/token, https://spacy.io/usage/processing-pipelines#custom-components-attributes

        """

        """Add custom attributes"""
        """Add full_shape attribute. Eg. 21.33 => dd.dd, esadDeweD23 => xxxxXxxxXdd"""

        def get_shape(token):
            full_shape = ""
            for i in token.text:
                if i.isdigit():
                    full_shape += "d"
                elif i.islower():
                    full_shape += "x"
                elif i.isupper():
                    full_shape += "X"
                else:
                    full_shape += i
            return full_shape

        spacy_token.set_extension("full_shape", getter=get_shape, force=True)

        def is_integer(token):
            pattern = re.compile('^[-+]?[0-9]+$')
            return bool(pattern.match(token.text))

        spacy_token.set_extension("is_integer", getter=is_integer, force=True)

        def is_decimal(token):
            pattern = re.compile('^[-+]?[0-9]+\.[0-9]+$')
            return bool(pattern.match(token.text))

        spacy_token.set_extension("is_decimal", getter=is_decimal, force=True)

        def is_ordinal(token):
            return token.orth_[-2:] in ['rd', 'st', 'th', 'nd']

        spacy_token.set_extension("is_ordinal", getter=is_ordinal, force=True)

        def is_mixed(token):
            if not token.is_title and not token.is_lower and not token.is_upper:
                return True
            else:
                return False

        spacy_token.set_extension("is_mixed", getter=is_mixed, force=True)

        """Add custom methods"""
        """Add get_prefix method. RETURN length N prefix"""

        def n_prefix(token, n):
            return token.text[:n]

        spacy_token.set_extension("n_prefix", method=n_prefix, force=True)

        """Add get_suffix method. RETURN length N suffix"""

        def n_suffix(token, n):
            return token.text[-n:]

        spacy_token.set_extension("n_suffix", method=n_suffix, force=True)

        return spacy_token

    @staticmethod
    def reconstruct_text(tokens: List[Token]) -> str:
        """
        Given a list of tokens, reconstruct the original text with as much fidelity as possible.

        Args:
            [tokens]:

        Returns: a string.

        """
        return "".join([x.text_with_ws for x in tokens])
